#include <gui/Context.h>
#include <gui/ApplicationWindow.h>
#include <gui/Viewer_2D.h>
#include <gui/ViewerInterface.h>

#include <scenegraph/OsgTransformObject.h>
#include <scenegraph/OsgDrawableObject.h>
#include <scenegraph/OsgDolfinObject.h>


#include <dolfin/mesh/Mesh.h>
#include <dolfin/mesh/MeshTopology.h>

#include <QDebug>
#include <QMdiArea>
#include <QMdiSubWindow>
#include <iostream>

//#define DEBUG_ADD_OBJECT
//#define DEBUG_SET_OBJECT
//#define DEBUG_CREATE_OSG_OBJECT
//#define DEBUG_REMOVE_OBJECT_FROM_SCENEGRAPH


//----------------------------------------------------------
Context::Context(ApplicationWindow* app) 
  : app_(*app), curr_viewer_(0)
//----------------------------------------------------------
{
}

//----------------------------------------------------------
Context::~Context()
//----------------------------------------------------------
{
}

//----------------------------------------------------------
ApplicationWindow* Context::app()
//----------------------------------------------------------
{
  return &app_;
}

//----------------------------------------------------------
const Context::ViewerList &Context::viewers() const
//----------------------------------------------------------
{
  return viewers_;
}

//----------------------------------------------------------
ViewerInterface *Context::viewer()
//----------------------------------------------------------
{
  return curr_viewer_;
}

//----------------------------------------------------------
void Context::setCurrentViewer(ViewerInterface* viewer)
//----------------------------------------------------------
{
  curr_viewer_ = viewer;
  
  QMenu* menu = app_.getViewerMenu();

  menu->clear();
  menu->addActions(viewer->getViewerMenu());

  emit viewerSelected(viewer);
}

//----------------------------------------------------------
QString Context::checkViewerName(const QString &name)
//----------------------------------------------------------
{
  QString n = name;
  int count = 1;
  while(viewers_.contains(n)) {	
    n = QString("%1 %2").arg(name).arg(count++);
  }
  return n;
}

//----------------------------------------------------------
void Context::addViewer(ViewerInterface* viewer, const QString name)
//----------------------------------------------------------
{
  //Find unique name for thewidget
  QString  ok_name = checkViewerName(name);

  
  QMdiSubWindow *s = app_.getMdiArea()->addSubWindow(viewer->get_widget());
  s->setWindowTitle(ok_name);
  s->setMinimumSize(320,200);
  s->show();

  if (nbViewers() == 0) {
    s->showMaximized();
  }
  else 
  {
    s->show();
    app_.getMdiArea()->tileSubWindows();
  }

  //Save the viewer
  viewers_[ok_name] = viewer;

  //Set as current viewer. 
  setCurrentViewer(viewer);
}

//----------------------------------------------------------
int Context::nbViewers() const
//----------------------------------------------------------
{
  return viewers_.count();
}

//----------------------------------------------------------
ViewerInterface *Context::getViewer(const QString &name)
//----------------------------------------------------------
{
  if (viewers_.contains(name))
    return viewers_[name];
  else
    return 0;
}

//----------------------------------------------------------
ViewerInterface *Context::getViewer(const int &id)
//----------------------------------------------------------
{
  ViewerList::iterator it = viewers_.begin();
  for (int i=0; it != viewers_.end(); ++it, ++i) {
    ViewerInterface *viewer = it.value();
    if (i==id)
      return viewer;
  }
  return 0;
}

//----------------------------------------------------------
void Context::deleteViewer(const QString &name)
//----------------------------------------------------------
{
  if (viewers_.contains(name))
    viewers_.remove(name);
}

//----------------------------------------------------------
void Context::updateViewers()
//----------------------------------------------------------
{
}

// //----------------------------------------------------------
// void Context::requestRedraw(bool update_gl)
// //----------------------------------------------------------
// {
//   //std::cerr << "Context::requestRedraw viewers.size()="<< viewers_.size() << std::endl;
  
//   ViewerList::iterator it = viewers_.begin();
//   for (int i=0; it != viewers_.end(); ++it, ++i) {
//     ViewerInterface *viewer = it.value();
//     viewer->requestRedraw();

//     //TODO: How to handle this
//     if (update_gl)
//       qDebug() << "Note: UpdateGL not implemented";
//       //viewer->updateGL(); 
//   }
// }

//----------------------------------------------------------
ObjectTree *Context::objectTree()
//----------------------------------------------------------
{
  return w_object_tree_;
}

//----------------------------------------------------------
void Context::setObjectTree(ObjectTree *w_tree)
//----------------------------------------------------------
{
  w_object_tree_ = w_tree;
}

//----------------------------------------------------------
ToolBoxDock* Context::toolBoxDock()
//----------------------------------------------------------
{
  return app_.getToolBoxDock();
}
//----------------------------------------------------------
MaterialSettings *Context::materialSettings()
//----------------------------------------------------------
{
  return material_settings_;
}

//----------------------------------------------------------
void Context::setMaterialSettings(MaterialSettings *mat)
//----------------------------------------------------------
{
  material_settings_ = mat;
}
//----------------------------------------------------------
SimResSettings *Context::simResSettings()
//----------------------------------------------------------
{
  return simres_settings_;
}
//----------------------------------------------------------
void Context::setSimResSettings(SimResSettings *simres)
//----------------------------------------------------------
{
  simres_settings_ = simres;
}
//----------------------------------------------------------
QAction *Context::undoAction()
//----------------------------------------------------------
{
  return a_undo_;
}

//----------------------------------------------------------
void Context::setUndoAction(QAction *action)
//----------------------------------------------------------
{
  a_undo_ = action;
}
//----------------------------------------------------------
void Context::importDolfinMesh(QString filename) 
//----------------------------------------------------------
{

  // Load the mesh
  DolfinObject *mesh = new DolfinObject(filename);

  ViewerInterface* viewer;

  // Create 2D or 3D viewer
  if (mesh->topology().dim() == 2) 
    viewer = new Viewer_2D(*this, *mesh, app());
  else 
    viewer = new Viewer(*this, *mesh, app());

  addViewer(viewer, filename);
  mesh->addViewer(viewer);

  //FIXME: calling setCurrentViewer gives a strange error.
  // Seems in the debugger as a signal is emitted...
  emit viewerSelected(curr_viewer_);
}
//----------------------------------------------------------
void Context::enableMeshFunctions(bool enable) 
//----------------------------------------------------------
{
  if (curr_viewer_)
    curr_viewer_->enableMeshFunctions(enable);
}
