#ifndef Gui_Context_h_Included             // -*- C++ -*-
#define Gui_Context_h_Included

#include <QObject>
#include <QMap>

class OsgTransformObject;
class ApplicationWindow;
class ObjectTree;
class ToolBoxDock;
class MaterialSettings;
class SimResSettings;
class TransformSettings;
class PreferencesDialog;
class QAction;
class Viewer;
class ViewerInterface;

/**
 * Global context of an application
 *
 * \ingroup gui
 *
 * \author Yvon Halbwachs
 * \author Benjamin Kehlet
 * \date 2011-01-28
 */

class Context : public QObject
{
  Q_OBJECT;
  
public:
  /** A list of viewers */
  typedef QMap<QString, ViewerInterface*> ViewerList;

  /** Empty constructor */
  Context(ApplicationWindow* app);
  
  /** Destroys the context. */
  virtual ~Context();
  
  /** Pointer to the main application window */
  ApplicationWindow *app();

  //+++++++++++++++++++++++
  // Viewer informations
  //+++++++++++++++++++++++
  
  /** Existing viewers */
  const ViewerList &viewers() const;
  
  /** Current active viewer */
  ViewerInterface *viewer();
  
  /** Specify current active viewer. */
  void setCurrentViewer(ViewerInterface* viewer);
  
  /** Creates a new viewer in the application */
  void addViewer(ViewerInterface* viewer, const QString name);
  
  /** Returns the number of opened viewers */
  int nbViewers() const;
  
  /** Returns the viewer associated to a name */
  ViewerInterface *getViewer(const QString &name);
  
  /** Returns the viewer number id.
   *  With id in the range [0, nbViewers()[
   */
  ViewerInterface *getViewer(const int &id);
  
  /** Removes a viewer from the list of viewers. */
  void deleteViewer(const QString &name);
  
  /** Update scenegraph for all viewers */
  void updateViewers();
  
  /** Request a redraw of all viewers.
      If "update_gl" is true, then updateGL is also
      called on each viewers (in addition to 
      requestRedraw).
  */
  //void requestRedraw(bool update_gl = false);
  
  /** Finds a unique name for a viewer */
  QString checkViewerName(const QString &name);

  /** Should this be here? */
  void importDolfinMesh(QString filename);
  
  
  //++++++++++++++++++++++++
  // Main widgets
  //++++++++++++++++++++++++

  //TODO: Move this to Application window as it is gui related?
  
  /** Object tree holding current loaded objects */
  ObjectTree *objectTree();
    
  /** Tool box dock with all tabs */
  //TODO: Move this to applicationwindow
  ToolBoxDock *toolBoxDock();
    
  /** Global material settings. */
  MaterialSettings *materialSettings();
  
  /** Global simres settings. */
  SimResSettings *simResSettings();
  
  /** Global undo action from the application edit menu */
  QAction *undoAction();

signals:
  /** Emitted when the current viewer has changed.
   *  This is emitted when setCurrentViewer is called. 
   */
  void viewerSelected(ViewerInterface* viewer);
  
public slots:
  void enableMeshFunctions(bool enable);

protected:
  
  /** Sets object tree widget */
  void setObjectTree(ObjectTree *w_tree);
    
  /** Sets the tool box dock with all tabs */
  //void setToolBoxDock(ToolBoxDock *w_tb);
    
  /** Sets global material settings. */
  void setMaterialSettings(MaterialSettings *mat);
  
  /** Sets global simres settings. */
  void setSimResSettings(SimResSettings *simres);
  
  /** Sets undo action from the application edit menu */
  void setUndoAction(QAction *action);

private:

  /** Pointer to global application window */
  ApplicationWindow& app_;  
  
  /** Object tree holding current loaded objects */
  ObjectTree *w_object_tree_;
  
  /** Tool box dock */
  ToolBoxDock *w_toolbox_;
  
  /** Global material settings */
  MaterialSettings* material_settings_;
  
  /** Global simres settings */
  SimResSettings* simres_settings_;
  
  /** Global transform settings */
  TransformSettings* transform_settings_;

  /** Global undo action */
  QAction *a_undo_;  

  /** 3d Viewers in the application */
  Context::ViewerList viewers_;
  
  /** Current project */
  //Project::Ref project_;
  
  /** Current active viewer */
  ViewerInterface *curr_viewer_;
  
  friend class ApplicationWindow;
  
};

#endif // Gui_Context_h_Included
