#include <gui/DolfinObject.h>
#include <gui/ViewerInterface.h>
#include <dolfin/mesh/Vertex.h>
#include <dolfin/mesh/Edge.h>
#include <dolfin/mesh/Face.h>
#include <dolfin/mesh/BoundaryMesh.h>

#include <QDebug>

//----------------------------------------------------------
DolfinObject::DolfinObject(QString filename)
  : dolfin::Mesh(filename.toStdString()), surface_mesh_(NULL)
//----------------------------------------------------------
{
  // Trigger generation of whatever is missing in the mesh.
  init(0);
  init(1);
  init(2);

  std::cout << "DolfinObject::initialize: Mesh has " 
            << size(0) << " vertices. " << std::endl;
  std::cout << "DolfinObject::initialize: Mesh has " 
            << size(1) << " edges. " << std::endl;
  std::cout << "DolfinObject::initialize: Mesh has " 
            << size(2) << " faces. " << std::endl;

  std::cout << "DolfinObject::initialize: Mesh topology dimension = " 
            << topology().dim() << std::endl;
  std::cout << "DolfinObject::initialize: Mesh geometry dimension = " 
            << geometry().dim() << std::endl;

  //=================================
  // Create boundary mesh
  //=================================

  std::cout << "DolfinObject::initialize: Creating boundary mesh from dolfin mesh" << std::endl;

  if(topology().dim() == 3) 
  {
    dolfin::BoundaryMesh* bndr_mesh = new dolfin::BoundaryMesh(*this);
    dolfin_vertices_ = &bndr_mesh->vertex_map();
    dolfin_faces_    = &bndr_mesh->cell_map();

    surface_mesh_ = bndr_mesh;

    // Trigger generation of edges in the surface mesh.
    surface_mesh_->init(1);

    // "Manually" generate the mapping between the edges in the full mesh and the surface mesh
    dolfin_edges_.init(*surface_mesh_, 1);

    for(dolfin::EdgeIterator v(*surface_mesh_); !v.end(); ++v)
    {
      const uint* entities = v->entities(0);
      uint  index = v->index();
      //uint  num_entities = v->num_entities(0);
      
      // Get the corresponding indices on the full mesh
      uint id1 = dolfin_vertices_->values()[entities[0]];
      uint id2 = dolfin_vertices_->values()[entities[1]];
      
      // Get the first vertex and iterate over the edges it's connected to
      dolfin::Vertex onevertex(*this, id1);
      
      for(dolfin::EdgeIterator e(onevertex); !e.end(); ++e)
      {
        const uint* full_entities = e->entities(0);
        uint  full_index    = e->index();
        //uint  full_num_entities = e->num_entities(0);
	
        // Compare indices. If they are the same we have a hit,
        // and add the mapping to the dolfin_edges_ map function.
        if(  ((id1 == full_entities[0]) && (id2 == full_entities[1]))
             || ((id1 == full_entities[1]) && (id2 == full_entities[0])))
        {
          dolfin_edges_[index] = full_index;
          break;
        }
      }
    }
  }
  else
  {
    // TODO: There must be a better way to do this
    std::cout << "DolfinObject::initialize: Mesh is 2D, mapping surface and mesh 1:1 " << std::endl;
    surface_mesh_ = this;
    
    dolfin_vertices_ = new dolfin::MeshFunction<unsigned int>(*surface_mesh_, 0);
    dolfin_faces_ = new dolfin::MeshFunction<unsigned int>(*surface_mesh_, 2);

    //dolfin_vertices_->init(*surface_mesh_, 0);
    dolfin_edges_.init(*surface_mesh_, 1);
    //dolfin_faces_->init(*surface_mesh_, 2);

    // Create 1:1 mappings for the vertices, edges and faces
    for(dolfin::VertexIterator v(*surface_mesh_); !v.end(); ++v)
      (*dolfin_vertices_)[v->index()] = v->index();
    for(dolfin::EdgeIterator v(*surface_mesh_); !v.end(); ++v)
      dolfin_edges_[v->index()] = v->index();
    for(dolfin::FaceIterator v(*surface_mesh_); !v.end(); ++v)
      (*dolfin_faces_)[v->index()] = v->index();
  }

  std::cout << "DolfinObject::initialize: Surface mesh has " << surface_mesh_->size(0) << " vertices. " << std::endl;
  std::cout << "DolfinObject::initialize: Surface mesh has " << surface_mesh_->size(1) << " edges. " << std::endl;
  std::cout << "DolfinObject::initialize: Surface mesh has " << surface_mesh_->size(2) << " faces. " << std::endl;

  std::cout << "DolfinObject::initialize: Boundary mesh mapping: vertices " << dolfin_vertices_->size()
            << " edges " << dolfin_edges_.size() << " cells " << dolfin_faces_->size() << " " << std::endl;

  //=================================
  // Initialize Vertex mesh function
  //=================================

  std::cout << "DolfinObject::initialize: Init vertex meshfunction: ";
  vertex_boundary_indicators_.init(*this, 0);
  std::cout << "Vertex meshfunction dim = " << vertex_boundary_indicators_.dim() 
            << " size = " << vertex_boundary_indicators_.size()<< std::endl;

  // Initialize all values to zero
  vertex_boundary_indicators_ = 0;

  vertex_undo_indicators_.init(*this, 0);
  vertex_undo_indicators_ = 0;

  //=================================
  // Initialize Edge mesh function
  //=================================

  std::cout << "DolfinObject::initialize:   Init edge meshfunction: ";
  edge_boundary_indicators_.init(*this, 1);
  std::cout << "  Edge meshfunction dim = " << edge_boundary_indicators_.dim() 
            << " size = " << edge_boundary_indicators_.size() << std::endl;

  // Initialize all values to zero
  edge_boundary_indicators_ = 0;

  edge_undo_indicators_.init(*this, 1);
  edge_undo_indicators_ = 0;

  //=================================
  // Initialize Face mesh function
  //=================================
  std::cout << "DolfinObject::initialize:   Init face meshfunction: ";
  face_boundary_indicators_.init(*this, 2);
  std::cout << "  Face meshfunction dim = " << face_boundary_indicators_.dim() 
            << " size = " << face_boundary_indicators_.size() << std::endl;

  // Initialize all values to zero
  face_boundary_indicators_ = 0;

  face_undo_indicators_.init(*this, 2);
  face_undo_indicators_ = 0;
}
//----------------------------------------------------------
DolfinObject::~DolfinObject()
//----------------------------------------------------------
{
  if(surface_mesh_ && surface_mesh_->topology().dim() == 2)
  {
    delete dolfin_vertices_;
    delete dolfin_faces_;
  }

  delete surface_mesh_;
}

//============================================================
// Get/Set & Load/Save boundary indicators
//============================================================


//==============
// Vertex
//==============
//----------------------------------------------------------
void DolfinObject::loadVertexBoundaryIndicators(QString filename)
//----------------------------------------------------------
{
  dolfin::File in(filename.toStdString());
  in >> vertex_boundary_indicators_;
}
//----------------------------------------------------------
void DolfinObject::saveVertexBoundaryIndicators(QString filename)
//----------------------------------------------------------
{
  dolfin::File out(filename.toStdString());
  out << vertex_boundary_indicators_;
}
//----------------------------------------------------------
int DolfinObject::getVertexBoundaryIndicator(uint id)
//----------------------------------------------------------
{
  uint dolfin_mesh_id = dolfin_vertices_->values()[id];
  return vertex_boundary_indicators_[dolfin_mesh_id];
}
//----------------------------------------------------------
void DolfinObject::setVertexBoundaryIndicator(uint id, int value, SELECT_MODE mode, bool redraw)
//----------------------------------------------------------
{

  switch (mode) {

  case SINGLE :
  {
    uint dolfin_mesh_id = dolfin_vertices_->values()[id];
    vertex_boundary_indicators_[dolfin_mesh_id] = value;
    
    if (redraw)
    {
      for (std::vector<ViewerInterface*>::iterator it = viewers.begin(); it < viewers.end(); it++)
      {
        (*it)->setVertexBoundaryIndicator(id, value);
      }
    }
    break;
  }

  case FILL :
  {
    uint dolfin_mesh_id = dolfin_vertices_->values()[id];
    int indicator = vertex_boundary_indicators_.values()[dolfin_mesh_id];
  
    if(indicator != value)
    {
      vertex_boundary_indicators_[dolfin_mesh_id] = value;
      dolfin::MeshEntity face(surface_mesh(), 0, id);
      for(dolfin::MeshEntityIterator v(face, 0); !v.end(); ++v)
        //recursive call
        setVertexBoundaryIndicator(v->index(), value, FILL, false);
    }

    if (redraw)
    {
      for (std::vector<ViewerInterface*>::iterator it = viewers.begin(); it < viewers.end(); it++)
      {
        (*it)->updateBoundaryIndicatorColors();
      }
    }

    break;
  }
  
  case ALL :
  {
    for(dolfin::VertexIterator v(surface_mesh()); !v.end(); ++v)
    {
      uint vertex_id = v->index();

      // Update the mesh function
      setVertexBoundaryIndicator(vertex_id, value, SINGLE, false);
    }
    
    if (redraw)
    {
      for (std::vector<ViewerInterface*>::iterator it = viewers.begin(); it < viewers.end(); it++)
      {
        (*it)->updateBoundaryIndicatorColors();
      }
    }

    break;
  }
  
  }

}
//==============
// Edge
//==============
//----------------------------------------------------------
void DolfinObject::loadEdgeBoundaryIndicators(QString filename)
//----------------------------------------------------------
{
  dolfin::File in(filename.toStdString());
  in >> edge_boundary_indicators_;
}
//----------------------------------------------------------
void DolfinObject::saveEdgeBoundaryIndicators(QString filename)
//----------------------------------------------------------
{
  dolfin::File out(filename.toStdString());
  out << edge_boundary_indicators_;
}
//----------------------------------------------------------
void DolfinObject::setEdgeBoundaryIndicator(uint id, int value, SELECT_MODE mode, bool redraw)
//----------------------------------------------------------
{


  switch (mode) {

  case SINGLE :
  {
   uint dolfin_mesh_id = dolfin_edges_.values()[id];
   edge_boundary_indicators_[dolfin_mesh_id] = value;

   for (std::vector<ViewerInterface*>::iterator it = viewers.begin(); it < viewers.end(); it++)
   {
     (*it)->setEdgeBoundaryIndicator(id, value);
   }

   break;
  }

  case FILL :
  {
    uint dolfin_mesh_id = dolfin_edges_.values()[id];
    int indicator = edge_boundary_indicators_.values()[dolfin_mesh_id];
  
    if(indicator != value)
    {
      edge_boundary_indicators_[dolfin_mesh_id] = value;
      dolfin::MeshEntity edge(surface_mesh(), 1, id);
      for(dolfin::MeshEntityIterator v(edge, 1); !v.end(); ++v)
      {
        //recursive call
        setEdgeBoundaryIndicator(v->index(), value, FILL, false);
      }
    }

    if (redraw)
    {
      for (std::vector<ViewerInterface*>::iterator it = viewers.begin(); it < viewers.end(); it++)
      {
        (*it)->updateBoundaryIndicatorColors();
      }
    }

    break;
  }
  
  case ALL :
  {
    for(dolfin::EdgeIterator v(surface_mesh()); !v.end(); ++v)
    {
      uint edge_id = v->index();

      // Update the mesh function
      setEdgeBoundaryIndicator(edge_id, value, SINGLE, false);
    }
    
    if (redraw)
    {
      for (std::vector<ViewerInterface*>::iterator it = viewers.begin(); it < viewers.end(); it++)
      {
        (*it)->updateBoundaryIndicatorColors();
      }
    }

    break;
  }
  
  }
}
//----------------------------------------------------------
int DolfinObject::getEdgeBoundaryIndicator(uint id)
//----------------------------------------------------------
{
  uint dolfin_mesh_id = dolfin_edges_.values()[id];
  return edge_boundary_indicators_[dolfin_mesh_id];
}
//==============
// Face
//==============
//----------------------------------------------------------
void DolfinObject::loadFaceBoundaryIndicators(QString filename)
//----------------------------------------------------------
{
  dolfin::File in(filename.toStdString());
  in >> face_boundary_indicators_;
}
//----------------------------------------------------------
void DolfinObject::saveFaceBoundaryIndicators(QString filename)
//----------------------------------------------------------
{
  dolfin::File out(filename.toStdString());
  out << face_boundary_indicators_;
}
//----------------------------------------------------------
void DolfinObject::setFaceBoundaryIndicator(uint id, int value, SELECT_MODE mode, bool redraw)
//----------------------------------------------------------
{
  switch (mode) {

  case SINGLE :
  {
    uint dolfin_mesh_id = dolfin_faces_->values()[id];
    face_boundary_indicators_[dolfin_mesh_id] = value;
    
    if (redraw)
    {
      for (std::vector<ViewerInterface*>::iterator it = viewers.begin(); it < viewers.end(); it++)
      {
        (*it)->setFaceBoundaryIndicator(id, value);
      }
    }
    break;
  }

  case FILL :
  {
    uint dolfin_mesh_id = dolfin_faces_->values()[id];
    int indicator = face_boundary_indicators_.values()[dolfin_mesh_id];
  
    if(indicator != value)
    {
      face_boundary_indicators_[dolfin_mesh_id] = value;
      dolfin::MeshEntity face(surface_mesh(), 2, id);
      for(dolfin::MeshEntityIterator v(face, 2); !v.end(); ++v)
        //recursive call
        setFaceBoundaryIndicator(v->index(), value, FILL, false);
    }

    if (redraw)
    {
      for (std::vector<ViewerInterface*>::iterator it = viewers.begin(); it < viewers.end(); it++)
      {
        (*it)->updateBoundaryIndicatorColors();
      }
    }

    break;
  }
  
  case ALL :
  {
    for(dolfin::FaceIterator v(surface_mesh()); !v.end(); ++v)
    {
      uint face_id = v->index();

      // Update the mesh function
      setFaceBoundaryIndicator(face_id, value, SINGLE, false);
    }
    
    if (redraw)
    {
      for (std::vector<ViewerInterface*>::iterator it = viewers.begin(); it < viewers.end(); it++)
      {
        (*it)->updateBoundaryIndicatorColors();
      }
    }

    break;
  }
  
  }
}
//----------------------------------------------------------
int DolfinObject::getFaceBoundaryIndicator(uint id)
//----------------------------------------------------------
{
  uint dolfin_mesh_id = dolfin_faces_->values()[id];
  return face_boundary_indicators_[dolfin_mesh_id];
}
//==================
// Undo system
//==================

//----------------------------------------------------------
void DolfinObject::undo()
//----------------------------------------------------------
{
  for(dolfin::VertexIterator v(*this); !v.end(); ++v)
    vertex_boundary_indicators_[v->index()] = vertex_undo_indicators_.values()[v->index()];
  for(dolfin::EdgeIterator v(*this); !v.end(); ++v)
    edge_boundary_indicators_[v->index()] = edge_undo_indicators_.values()[v->index()];
  for(dolfin::FaceIterator v(*this); !v.end(); ++v)
    face_boundary_indicators_[v->index()] = face_undo_indicators_.values()[v->index()];
  
  std::cout << "DolfinObject::undo: Meshfunctions reset to previous version." << std::endl;
}

//----------------------------------------------------------
void DolfinObject::updateUndoBuffers()
//----------------------------------------------------------
{
  for(dolfin::VertexIterator v(*this); !v.end(); ++v)
    vertex_undo_indicators_[v->index()] = vertex_boundary_indicators_.values()[v->index()];
  for(dolfin::EdgeIterator v(*this); !v.end(); ++v)
    edge_undo_indicators_[v->index()] = edge_boundary_indicators_.values()[v->index()];
  for(dolfin::FaceIterator v(*this); !v.end(); ++v)
    face_undo_indicators_[v->index()] = face_boundary_indicators_.values()[v->index()];
}
//----------------------------------------------------------
void DolfinObject::addViewer(ViewerInterface* v)
//----------------------------------------------------------
{
  qDebug() << "DolfinObject::addViewer";
  viewers.push_back(v);
}
