#include "ViewerSettingsDialog.h"

#include <QtGui>
#include <gui/ColorButton.h>
#include <gui/ApplicationWindow.h>
#include <gui/Viewer.h>

//----------------------------------------------------------
ViewerSettingsDialog::ViewerSettingsDialog(Viewer* viewer, ApplicationWindow* parent) 
  :QDialog(parent),
   viewer(viewer)
//----------------------------------------------------------
{


  /************************* Setup ui ******************************/
  verticalLayout = new QVBoxLayout(this);

  primary_background_color = new ColorButton(this);
  primary_background_color->setColor(viewer->getBackgroundColor());
  primary_background_color->setLabel(tr("Background around objects"));  
  verticalLayout->addWidget(primary_background_color);
  
  ambient_light_color = new ColorButton(this);
  ambient_light_color->setLabel(tr("Ambient light color"));
  ambient_light_color->setColor(QColor(0, 0, 0));
  verticalLayout->addWidget(ambient_light_color);
  
  diffuse_light_color = new ColorButton(this);
  diffuse_light_color->setLabel(tr("Diffuse light color"));
  diffuse_light_color->setColor(QColor(188, 188, 188));
  verticalLayout->addWidget(diffuse_light_color);
  
  specular_light_color = new ColorButton(this);
  specular_light_color->setLabel(tr("Specular light color"));
  specular_light_color->setColor(QColor(235, 235, 235));
  verticalLayout->addWidget(specular_light_color);
  
  window_size_combo = new QComboBox(this);
  window_size_combo->setEditable(false);
  window_size_combo->insertItems(0, QStringList()
	 << tr("Expanding")        << tr("1280x1024")       << tr("1024x768")
         << tr("800x600")          << tr("640x400")         << tr("512x512")
         << tr("320x200")          << tr("256x256")         << tr("768x576 PAL Analog")
         << tr("720x576 PAL CCIR") << tr("720x432 PAL 16:9")<< tr("720x486 NTSC CCIR")
         << tr("646x486 NTSC Analog")
        );
  verticalLayout->addWidget(window_size_combo);

  
  spacer = new QSpacerItem(20, 20, QSizePolicy::Minimum, QSizePolicy::Expanding);
  verticalLayout->addItem(spacer);

  closeButton = new QPushButton(tr("Close"));
  verticalLayout->addWidget(closeButton);

 
  /*************************** Connect widgets ********************************/
  
  // Window size menu
  connect(window_size_combo, SIGNAL(activated(const QString&)),
	  this, SLOT(changeWindowSize(const QString&)));



  // Background color
  connect(primary_background_color, SIGNAL(colorChanged(const QColor& )),
	  viewer, SLOT(setBackgroundColor(const QColor&)));

  // Ambient light color
  connect(ambient_light_color, SIGNAL(colorChanged(const QColor&)),
	  this, SLOT(changeLightAmbient(const QColor&)));

  // Diffuse light color
  connect(diffuse_light_color, SIGNAL(colorChanged(const QColor&)),
	  this, SLOT(changeLightDiffuse(const QColor&)));

  // Specular light color
  connect(specular_light_color, SIGNAL(colorChanged(const QColor&)),
	  this, SLOT(changeLightSpecular(const QColor&)));


  connect(closeButton, SIGNAL(clicked()),
	  this, SLOT(accept()));
  
}

//----------------------------------------------------------
void ViewerSettingsDialog::changeWindowSize(const QString& arg)
//----------------------------------------------------------
{
  QRegExp regexp("(\\d+)x(\\d+)");
  
  int pos = regexp.indexIn(arg);
  if(pos>=0) {
    int w = regexp.cap(1).toInt();
    int h = regexp.cap(2).toInt();
    
    //std::cerr << "Changing size to " << w << " " << h << std::end;
    viewer->setFixedSize(w,h);
  }
  else {
    //std::cerr << "No match, pos=" << pos << std::end;
    viewer->setMinimumSize(320,200);
    viewer->setMaximumSize(32767,32767);
  }
  viewer->updateGL();
}

//----------------------------------------------------------
void ViewerSettingsDialog::changeLightAmbient(const QColor& c)
//----------------------------------------------------------
{
  osg::Light* light = viewer->getLight();
  
  if(light) {
    osg::Vec4 vec(c.red()/255.0,
		  c.green()/255.0,
		  c.blue()/255.0,
		  c.alpha()/255.0);
    
    light->setAmbient(vec);
  }
}

//----------------------------------------------------------
void ViewerSettingsDialog::changeLightDiffuse(const QColor& c)
//----------------------------------------------------------
{
  osg::Light* light = viewer->getLight();
  
  if(light) {
    osg::Vec4 vec(c.red()/255.0,
		  c.green()/255.0,
		  c.blue()/255.0,
		  c.alpha()/255.0);
      
    light->setDiffuse(vec);
  }
}

//----------------------------------------------------------
void ViewerSettingsDialog::changeLightSpecular(const QColor& c)
//----------------------------------------------------------
{

  osg::Light* light = viewer->getLight();
  
  if(light) {
    osg::Vec4 vec(c.red()/255.0,
		  c.green()/255.0,
		  c.blue()/255.0,
		  c.alpha()/255.0);
    
    light->setSpecular(vec);
  }
}
