#include <gui/ApplicationWindow.h>
#include <gui/Viewer_2D.h>
#include <gui/Scene_2D.h>
#include <gui/Vertex_2D.h>
#include <gui/Edge_2D.h>
#include <gui/Axes_2D.h>
#include <gui/DolfinObject.h>
#include <gui/MeshFunctionsToolbar.h>
#include <gui/VisualizationDialog_2D.h>

#include <dolfin/mesh/Mesh.h>
#include <dolfin/mesh/Vertex.h>
#include <dolfin/mesh/Edge.h>
#include <dolfin/mesh/Point.h>

#include <QDebug>
#include <QGraphicsEllipseItem>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsItem>
#include <QWheelEvent>
#include <QMouseEvent>


#include <float.h>
#include <iostream>

#define DRAW_VERTICES 1
#define DRAW_EDGES 1
//----------------------------------------------------------
Viewer_2D::Viewer_2D(Context& context, DolfinObject& mesh, QWidget* parent) 
//----------------------------------------------------------
  : context(context), mesh(mesh), showAxes(true)
  
{
  initViewerMenu();

  scene = new Scene_2D(mesh, this);
  setScene(scene);

  setRenderHint(QPainter::Antialiasing);

  if (DRAW_VERTICES)
  {
    for(dolfin::VertexIterator it(mesh.surface_mesh()); !it.end(); ++it) 
    {
      const dolfin::Point p = it->point();
      scene->addItem(new Vertex_2D(p[0], p[1], it->index()));
    }
  }


  double x_min = DBL_MAX, x_max = -DBL_MAX, y_min = DBL_MAX, y_max = -DBL_MAX;

  for (dolfin::EdgeIterator it(mesh.surface_mesh()); !it.end(); ++it) 
  {
    const uint* vertices = it->entities(0);
    assert(vertices);

    const dolfin::Vertex v0(mesh, vertices[0]);
    const dolfin::Vertex v1(mesh, vertices[1]);

    const dolfin::Point p0 = v0.point();
    x_min = std::min(x_min, p0[0]); x_max = std::max(x_max, p0[0]);
    y_min = std::min(y_min, p0[1]); y_max = std::max(y_max, p0[1]);

    const dolfin::Point p1 = v1.point();
    x_min = std::min(x_min, p1[0]); x_max = std::max(x_max, p1[0]);
    y_min = std::min(y_min, p1[1]); y_max = std::max(y_max, p1[1]);

    if (DRAW_EDGES)
    {
      Edge_2D* e = new Edge_2D(p0[0], p0[1], p1[0], p1[1], it->index());
      e->setZValue(-1);
      scene->addItem(e);
    }
  }

  // TODO: Draw axes here
  axes = new Axes_2D(x_min, x_max, y_min, y_max);
  axes->setVisible(showAxes);
  scene->addItem(axes);

  setSceneRect(x_min, y_min, x_max - x_min, y_max - y_min);

  // Flip y axis
  scale(1, -1);

  // Zoom to fit in widget, but keep aspect ratio
  fitInView(x_min, y_min, x_max - x_min, y_max - y_min, Qt::KeepAspectRatio);

  // Enable mouseMoveEvent, otherwise events are fired only when a mouse 
  // is pressed
  setMouseTracking(true);

  setDragMode(ScrollHandDrag);
}
//----------------------------------------------------------
Viewer_2D::~Viewer_2D()
//----------------------------------------------------------
{
}
//----------------------------------------------------------
void Viewer_2D::wheelEvent(QWheelEvent *event) 
//----------------------------------------------------------
{
  double numDegrees = -event->delta() / 8.0;
  double numSteps = numDegrees / 15.0;
  double factor = std::pow(1.125, numSteps);
  scale(factor, factor);
}
//----------------------------------------------------------
void Viewer_2D::viewCenter(){}
//----------------------------------------------------------

//----------------------------------------------------------
void Viewer_2D::updateBoundaryIndicatorColors()
//----------------------------------------------------------
{
  scene->updateBoundaryIndicatorColors();
}
//----------------------------------------------------------
void Viewer_2D::saveImage(QString filename){}
//----------------------------------------------------------

//----------------------------------------------------------
void Viewer_2D::undo(){}
//----------------------------------------------------------

//----------------------------------------------------------
void Viewer_2D::enableMeshFunctions(bool enable)
//----------------------------------------------------------
{
  scene->enableMarking(enable, context.app()->meshfunctions_toolbar);
}

void Viewer_2D::setVertexBoundaryIndicator(uint id, int value){}
void Viewer_2D::setEdgeBoundaryIndicator(uint id, int value){}
void Viewer_2D::setFaceBoundaryIndicator(uint id, int value){}
//----------------------------------------------------------
void Viewer_2D::initViewerMenu()
//----------------------------------------------------------
{  
  QAction* vis_action = new QAction(QString("Visualization"), this);
  connect(vis_action, SIGNAL(triggered()),
          this, SLOT(visualization()));
  viewer_menu.append(vis_action);

}
//----------------------------------------------------------
void Viewer_2D::visualization()
//----------------------------------------------------------
{
  VisualizationDialog_2D d(this, context.app());
  d.exec(); 
}
//----------------------------------------------------------
void Viewer_2D::showAxesChanged(bool show)
//----------------------------------------------------------
{
  axes->setVisible(show);
}

