#include "VisualizationDialog.h"
#include <gui/ApplicationWindow.h>
#include <gui/Viewer.h>
#include <gui/SimResSettings.h>
#include <gui/MaterialSettings.h>
#include <gui/ColorButton.h>
#include <gui/FloatSlider.h>

#include <QtGui>
#include <iostream>

//----------------------------------------------------------
VisualizationDialog::VisualizationDialog(Viewer* viewer, ApplicationWindow* parent)
  : QDialog(parent),
    simres_settings(viewer->getSimres_settings()),
    material_settings(viewer->getMaterial_settings())
//----------------------------------------------------------
{
  // Create the tab widget
  QVBoxLayout* layout = new QVBoxLayout(this);
  tabs = new QTabWidget(this);
  tabs->setTabShape(QTabWidget::Rounded);
  tabs->setElideMode(Qt::ElideRight);
  tabs->setUsesScrollButtons(false);


  //Coloring page
  coloring_page = new QWidget();
  coloring_page_layout = new QVBoxLayout(coloring_page);
  coloring_page_layout->setSpacing(6);
  coloring_page_layout->setMargin(6);

  render_mode_groupbox = new QGroupBox("Render mode", coloring_page);
  render_mode_groupbox->setFlat(false);
  render_mode_groupbox->setCheckable(false);
  
  render_mode = new QButtonGroup();  
  form_layout = new QFormLayout(render_mode_groupbox);
  form_layout->setSpacing(6);
  form_layout->setMargin(6);

  smooth_button = new QRadioButton(render_mode_groupbox);
  form_layout->addRow("Smooth", smooth_button);
  render_mode->addButton(smooth_button, SimResState::SMOOTH);
  
  wireframe_button = new QRadioButton(render_mode_groupbox);
  form_layout->addRow("Wireframe", wireframe_button);
  render_mode->addButton(wireframe_button, SimResState::WIREFRAME);
  
  flat_button = new QRadioButton(render_mode_groupbox);
  form_layout->addRow("Flat", flat_button);
  render_mode->addButton(flat_button, SimResState::FLAT);  

  wire_and_flat_button = new QRadioButton(render_mode_groupbox);
  form_layout->addRow("Wire and flat", wire_and_flat_button);
  render_mode->addButton(wire_and_flat_button, SimResState::WIRE_AND_FLAT);  

  points_button = new QRadioButton(render_mode_groupbox);
  form_layout->addRow("Points", points_button);
  render_mode->addButton(points_button, SimResState::POINTS);  

  wire_and_smooth_button = new QRadioButton(render_mode_groupbox);
  form_layout->addRow("Wire and smooth", wire_and_smooth_button);
  render_mode->addButton(wire_and_smooth_button, SimResState::WIRE_AND_SMOOTH);
  
  // NOTE: Since the id of the buttons are initialized with 
  // the values of SimResState::RENDER_MODE we can send the id of the
  // button directly.
  int checked_button_id = simres_settings->currState().getRenderMode(); 
  render_mode->button(checked_button_id)->setChecked(true);

  coloring_page_layout->addWidget(render_mode_groupbox);
  
  base_color = new ColorButton(coloring_page);
  base_color->setLabel("Base color");
  base_color->setColor(simres_settings->currState().getBaseColor());
  coloring_page_layout->addWidget(base_color);
  
  back_color = new ColorButton(coloring_page);
  back_color->setLabel("Back color");
  back_color->setColor(simres_settings->currState().getBackColor());
  coloring_page_layout->addWidget(back_color);
  
  highlight_color = new ColorButton(coloring_page);
  highlight_color->setLabel("Highlight color");
  highlight_color->setColor(simres_settings->currState().getHighlightColor());
  coloring_page_layout->addWidget(highlight_color);

  show_highlight = new QCheckBox("Show highlight", this);
  show_highlight->setChecked(simres_settings->currState().getHighlight());
  coloring_page_layout->addWidget(show_highlight);
  
  show_axes = new QCheckBox("Show axes", this);
  show_axes->setChecked(simres_settings->currState().getShowAxes());
  coloring_page_layout->addWidget(show_axes);

  show_plane = new QCheckBox("Show ground plane", this);
  show_plane->setChecked(simres_settings->currState().getShowPlane());
  coloring_page_layout->addWidget(show_plane);

  wireframe_color = new ColorButton(coloring_page);
  wireframe_color->setLabel("Wireframe color");
  wireframe_color->setColor(simres_settings->currState().getWireframeColor());
  coloring_page_layout->addWidget(wireframe_color);
  
  
  coloring_page_layout->addItem(new QSpacerItem(20, 21, QSizePolicy::Minimum, QSizePolicy::Expanding));
  
  tabs->addTab(coloring_page, "Coloring");

  // Material tab
  material_page = new QWidget();
  material_layout = new QVBoxLayout(material_page);
  material_layout->setSpacing(6);
  material_layout->setMargin(6);

  //    Transparency
  transparency_slider = new FloatSlider(material_page);
  transparency_slider->setLabel("Transparency");
  transparency_slider->setMinValue(0);
  transparency_slider->setMaxValue(1);
  transparency_slider->setGranul(0.01);
  material_layout->addWidget(transparency_slider);
  
  //    Ambient
  ambient_slider = new FloatSlider(material_page);
  ambient_slider->setLabel("Ambient");
  ambient_slider->setMinValue(0);
  ambient_slider->setMaxValue(1);
  ambient_slider->setGranul(0.01);
  ambient_slider->setCurrentValue(0.4);
  material_layout->addWidget(ambient_slider);
  
  //    Diffuse
  diffuse_slider = new FloatSlider(material_page);
  diffuse_slider->setLabel("Diffuse");
  diffuse_slider->setMinValue(0);
  diffuse_slider->setMaxValue(1);
  diffuse_slider->setGranul(0.01);
  diffuse_slider->setCurrentValue(0.56);
  material_layout->addWidget(diffuse_slider);

  //    Emission
  emission_slider = new FloatSlider(material_page);
  emission_slider->setLabel("Emission");
  emission_slider->setMinValue(0);
  emission_slider->setMaxValue(1);
  emission_slider->setGranul(0.01);
  emission_slider->setCurrentValue(0.04);
  material_layout->addWidget(emission_slider);
  
  //    Specular
  specular_slider = new FloatSlider(material_page);
  specular_slider->setLabel("Specular");
  specular_slider->setMinValue(0);
  specular_slider->setMaxValue(1);
  specular_slider->setGranul(0.01);
  specular_slider->setCurrentValue(0);
  material_layout->addWidget(specular_slider);
  
  //    Shininess
  shininess_slider = new FloatSlider(material_page);
  shininess_slider->setLabel("Shininess");
  shininess_slider->setMinValue(0);
  shininess_slider->setMaxValue(128);
  shininess_slider->setGranul(1.0);
  shininess_slider->setCurrentValue(0);
  material_layout->addWidget(shininess_slider);
  
  material_layout->addItem(new QSpacerItem(20, 40, QSizePolicy::Minimum, QSizePolicy::Expanding));

  tabs->addTab(material_page, "Material");  

  
  advanced_page = new QWidget();
  advanced_layout = new QVBoxLayout(advanced_page);
  advanced_layout->setSpacing(6);
  advanced_layout->setMargin(6);

  //   Transparent sorting
  transparent_group = new QGroupBox(advanced_page);
  transparent_layout = new QHBoxLayout(transparent_group);
  transparent_layout->setSpacing(6);
  transparent_layout->setMargin(6);

  sort_mode_combo = new QComboBox(transparent_group);
  sort_mode_combo->addItem("Automatic depth sort OFF");
  sort_mode_combo->addItem("Re-sort every frame");
  transparent_layout->addWidget(sort_mode_combo);

  depth_sort_button = new QPushButton("Sort now", transparent_group);
  transparent_layout->addWidget(depth_sort_button);

  advanced_layout->addWidget(transparent_group);

  //   Culling
  culling_panel = new QGroupBox("Culling", advanced_page);
  culling_layout = new QHBoxLayout(culling_panel);
  culling_layout->setSpacing(6);
  culling_layout->setMargin(6);

  culling_button_group = new QButtonGroup;

  cull_mode_off = new QRadioButton("None", culling_panel);  
  culling_layout->addWidget(cull_mode_off);
  culling_button_group->addButton(cull_mode_off);
  
  cull_mode_back = new QRadioButton("Back", culling_panel);
  culling_layout->addWidget(cull_mode_back);
  culling_button_group->addButton(cull_mode_back);
  
  cull_mode_front = new QRadioButton("Front", culling_panel);
  culling_layout->addWidget(cull_mode_front);
  culling_button_group->addButton(cull_mode_front);
  
  advanced_layout->addWidget(culling_panel);
  
  //   Light model
  light_model_panel = new QGroupBox("Light model", advanced_page);
  light_model_layout = new QVBoxLayout(light_model_panel);
  light_model_layout->setSpacing(6);
  light_model_layout->setMargin(6);

  two_sided_checkbox = new QCheckBox("Two sided lighting", light_model_panel);
  light_model_layout->addWidget(two_sided_checkbox);
  
  advanced_layout->addWidget(light_model_panel);

  //  Points
  points_group = new QGroupBox("Points", advanced_page);
  points_layout = new QVBoxLayout(points_group);
  points_layout->setSpacing(6);
  points_layout->setMargin(6);
  point_size = new FloatSlider(points_group);
  point_size->setLabel("Size");
  point_size->setMinValue(1);
  point_size->setMaxValue(20);
  point_size->setGranul(0.01);
  points_layout->addWidget(point_size);

  advanced_layout->addWidget(points_group);
  
  
  line_group = new QGroupBox("Lines", advanced_page);
  line_layout = new QVBoxLayout(line_group);
  line_layout->setSpacing(6);
  line_layout->setMargin(6);

  line_width = new FloatSlider(line_group);
  line_width->setLabel("Width");
  line_width->setMinValue(1);
  line_width->setMaxValue(20);
  line_width->setGranul(0.01);
  line_layout->addWidget(line_width);
  
  advanced_layout->addWidget(line_group);

  //   Contours
  contours = new QGroupBox("Contours", advanced_page);
  contours->setCheckable(true);
  contours->setChecked(false);
  contour_layout = new QHBoxLayout(contours);
  contour_layout->setSpacing(6);
  contour_layout->setMargin(6);

  contour_layout->addWidget(new QLabel("Number of planes", contours));
  
  contour_nb = new QSpinBox(contours);
  contour_layout->addWidget(contour_nb);
  
  advanced_layout->addWidget(contours);
  
  //   Extras
  extras_group = new QGroupBox("Extra", advanced_page);
  extras_layout = new QVBoxLayout(extras_group);
  extras_layout->setSpacing(6);
  extras_layout->setMargin(6);

  show_bounding_sphere = new QCheckBox("Show bounding sphere", extras_group);
  extras_layout->addWidget(show_bounding_sphere);
  
  show_triangle_normals = new QCheckBox("Show triangle normals", extras_group);
  extras_layout->addWidget(show_triangle_normals);
  
  
  advanced_layout->addWidget(extras_group);
  
  advanced_layout->addItem(new QSpacerItem(20, 20, QSizePolicy::Minimum, QSizePolicy::Expanding));
  
  tabs->addTab(advanced_page, "Advanced");
    
  layout->addWidget(tabs);

  // Close button
  close_button = new QPushButton("Close");
  layout->addWidget(close_button);


  // Setup connections

  //  Connections to SimResSettings
  connect(base_color, SIGNAL(colorChanged(QRgb)),
          simres_settings, SLOT(simresBaseColorChanged(QRgb)));

  connect(back_color, SIGNAL(colorChanged(QRgb)),
          simres_settings, SLOT(simresBackColorChanged(QRgb)));
  
  connect(wireframe_color, SIGNAL(colorChanged(QRgb)),
          simres_settings, SLOT(simresWireframeColorChanged(QRgb)));

  connect(highlight_color, SIGNAL(colorChanged(QRgb)),
          simres_settings, SLOT(simresHighlightColorChanged(QRgb)));
  
  connect(show_highlight, SIGNAL(toggled(bool)),
          simres_settings, SLOT(showHighlightChanged(bool)));

  connect(show_axes, SIGNAL(toggled(bool)),
          simres_settings, SLOT(showAxesChanged(bool)));

  connect(show_plane, SIGNAL(toggled(bool)),
          simres_settings, SLOT(showPlaneChanged(bool)));

  connect(render_mode, SIGNAL(buttonClicked(int)),
          simres_settings, SLOT(simresRenderModeChanged(int)));
  
  connect(culling_button_group, SIGNAL(buttonClicked(int)),
          simres_settings, SLOT(simresCullModeChanged(int)));
  
  connect(two_sided_checkbox, SIGNAL(toggled(bool)),
          simres_settings, SLOT(simresTwoSidedLightingChanged(bool)));
  
  connect(transparency_slider, SIGNAL(valueChanged(double)),
          simres_settings, SLOT(simresTransparancyChanged(double)));
  
  connect(depth_sort_button, SIGNAL(clicked()),
          simres_settings, SLOT(simresDepthSortButtonPressed()));
  
  connect(sort_mode_combo, SIGNAL(activated(const QString&)),
          simres_settings, SLOT(simresSortModeSelected(const QString&)));
	      
  connect(point_size, SIGNAL(valueChanged(double)),
          simres_settings, SLOT(pointSizeChanged(double)));

  connect(line_width, SIGNAL(valueChanged(double)),
          simres_settings, SLOT(lineWidthChanged(double)));

  connect(contours, SIGNAL(toggled(bool)),
          simres_settings, SLOT(contoursChanged(bool)));

  connect(contour_nb, SIGNAL(valueChanged(int)),
          simres_settings, SLOT(contourNbChanged(int)));

  connect(show_bounding_sphere, SIGNAL(clicked(bool)),
          simres_settings, SLOT(showBoundingSphereChanged(bool)));

  connect(show_triangle_normals, SIGNAL(clicked(bool)),
          simres_settings, SLOT(showTriangleNormalsChanged(bool)));

  //    Connections to MaterialSettings
  connect(diffuse_slider, SIGNAL(valueChanged(double)),
          material_settings, SLOT(materialDiffuseChanged(double)));
  
  connect(ambient_slider, SIGNAL(valueChanged(double)),
          material_settings, SLOT(materialAmbientChanged(double)));
  
  connect(emission_slider, SIGNAL(valueChanged(double)),
          material_settings, SLOT(materialEmissionChanged(double)));
  
  connect(specular_slider, SIGNAL(valueChanged(double)),
          material_settings, SLOT(materialSpecularChanged(double)));
  
  connect(shininess_slider, SIGNAL(valueChanged(double)),
          material_settings, SLOT(materialShininessChanged(double)));
  

  connect(close_button, SIGNAL(clicked()),
          this, SLOT(accept()));

}
