#include <osgQt/RenderArea.h>
#include <osgQt/TrackballManipulator.h>

#include <osg/MatrixTransform>

#include <QImage>
#include <QTimer>


// This should be easier than this:
osg::Timer RenderArea::timer_;
const osg::Timer_t RenderArea::initialTick_ = RenderArea::timer_.tick();
  
//----------------------------------------------------------
RenderArea::RenderArea(QWidget* parent, const char* name, const QGLWidget* shareWidget, WindowFlags f):
  AdapterWidget(parent, name, shareWidget, f),
  fovy_(50.0),
  qtimer_(0),
  vp_w_(-1),
  vp_h_(-1),
  redrawPolicy_(AS_NEEDED),
  projectionMode_(PM_Perspective),
  redraw_time_(10),
  needContinuousUpdate_(false),
  needRedraw_(false),
  ignore_redraw_(false),
  z_near_(1.0),
  z_far_(10000.0)
  //----------------------------------------------------------
{
}
  
//----------------------------------------------------------
RenderArea::~RenderArea()
//----------------------------------------------------------
{
  if (qtimer_) 
    qtimer_->stop();
}

//----------------------------------------------------------
void RenderArea::initializeGL()
//----------------------------------------------------------
{

  // check that the openGL context is ok (inherited from QGLWidget)
  if(!isValid()){
    return;
  }
  
  getCamera()->setViewport(new osg::Viewport(0,0,width(),height()));
  getCamera()->setGraphicsContext(getGraphicsWindow());
    
  setProjection();
  
  setThreadingModel(osgViewer::Viewer::SingleThreaded);
  
  glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE);
  glColorMaterial(GL_BACK,GL_AMBIENT_AND_DIFFUSE);
  glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
  glColorMaterial(GL_FRONT_AND_BACK,GL_SPECULAR);
  glEnable(GL_COLOR_MATERIAL);
  
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,1);
  
#ifndef WIN32
  glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
#endif
  
  getCamera()->setViewport(0,0,width(),height());
  
  // Set the appropriate aspect ratio on the scene view's projection matrix
  setProjection();
}

//----------------------------------------------------------
void RenderArea::resizeGL(int w,int h)
//----------------------------------------------------------
{
  //cerr << "RenderArea::resizeGL w=" << w << ", h=" << h << endl; 
  
  AdapterWidget::resizeGL(w,h);
  
  if(!isValid()){
    return;
  }
  
  int x = 0;
  int y = 0;
  int width = w;
  int height = h;
  
  if(vp_w_ > 0 && vp_h_ > 0) {
    width  = vp_w_;
    height = vp_h_;
    x = (w - width)/2;
    y = (h - height)/2;
  }
  
  
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // Clear the old area if we get a smaller viewport
  
  getCamera()->setViewport(x, y, width, height);
  
  // Set the appropriate aspect ratio on the scene view's projection matrix
  setProjection();
  
  emit(viewerResized(this, width, height));
}

//----------------------------------------------------------
void RenderArea::updateModelView()
//----------------------------------------------------------
{
  // Update the model view on the scene view.
/*  if(matrixManipulator_.get())
    {
      osg::Matrix invmat = matrixManipulator_->getInverseMatrix();
      getCamera()->setViewMatrix(invmat);
    }*/
}

//----------------------------------------------------------
void RenderArea::processAdaptedEvents()
//----------------------------------------------------------
{
  //cerr << "RenderArea::processAdaptedEvents()" << endl;
  eventTraversal();
}
  
//----------------------------------------------------------
void RenderArea::paintGL()
//----------------------------------------------------------
{
  // This makes disapear cessna.osg!!
  //glDisableClientState(GL_VERTEX_ARRAY);


  //cerr << "RenderArea::paintGL()" << endl;
    
  // Reset needRedraw_ flag.
  needRedraw_ = false; 
    
  if(getSceneData() == NULL) return;
    
  if(!isValid()) return;
    
  emit enteringPaintGL();
    
  static osg::Timer_t enter_time;
  enter_time=timer_.tick();
    
  osg::FrameStamp* framestamp = getViewerFrameStamp();
  framestamp->setFrameNumber(framestamp->getFrameNumber()+1);
  framestamp->setReferenceTime(timer_.delta_s(initialTick_,enter_time));
    
  //osg::Timer_t before_update=timer_.tick();
    
  // set the view
  updateModelView();
    
  // draw frame
  // inherited from osgViewer::ViewerBase    
  frame();
    
  emit leavingPaintGL();
}
  
//----------------------------------------------------------
//void RenderArea::setMatrixManipulator(osgGA::MatrixManipulator* manip)
//----------------------------------------------------------
//{
//  matrixManipulator_ = manip;
  //addEventHandler(manip);
//  setCameraManipulator(manip);
//}

//----------------------------------------------------------
//osgGA::MatrixManipulator* RenderArea::getMatrixManipulator() 
//----------------------------------------------------------
//{
//  return matrixManipulator_.get();
//}

//----------------------------------------------------------
TrackballManipulator* RenderArea::getTrackballManipulator()
//----------------------------------------------------------
{
  return dynamic_cast<TrackballManipulator*>(getCameraManipulator());
}
  
//----------------------------------------------------------
const osg::Matrix RenderArea::computeMVPW() const
//----------------------------------------------------------
{
  osg::Matrix matrix(getCamera()->getViewMatrix() * getCamera()->getProjectionMatrix());
    
  if (getCamera()->getViewport())
    matrix.postMult(getCamera()->getViewport()->computeWindowMatrix());
  else
    osg::notify(osg::WARN)<<"osg::Matrix RenderArea::computeMVPW() - error no viewport attached to SceneView, coords will be computed inccorectly."<<std::endl;
    
  return matrix;
}
  
//----------------------------------------------------------
void RenderArea::projectWindowXYIntoObject(int x, int y, osg::Vec3& near_point, osg::Vec3& far_point) const
//----------------------------------------------------------
{
  osg::Matrix inverseMVPW;
  inverseMVPW.invert(computeMVPW());
    
  near_point = osg::Vec3(x,y,0.0f)*inverseMVPW;
  far_point = osg::Vec3(x,y,1.0f)*inverseMVPW;
}
  
//----------------------------------------------------------
void RenderArea::setRedrawPolicy(const RedrawPolicy rp)
//----------------------------------------------------------
{
  //if(getRedrawPolicy()==rp) return;
    
  redrawPolicy_=rp;
    
  if(redrawPolicy_==AS_NEEDED)
    {
      if(getNeedContinuousUpdate())
	{
	  if( !timer()->isActive() ) timer()->start(redraw_time_);
	}
      else
	{
	  if( timer()->isActive() ) timer()->stop();
	}
    }
  else if(redrawPolicy_==CONSTANT)
    {
      if( !timer()->isActive() ) timer()->start(redraw_time_);
    }
}
  
//----------------------------------------------------------
RenderArea::RedrawPolicy RenderArea::getRedrawPolicy() const 
//----------------------------------------------------------
{
  return redrawPolicy_;
}

//----------------------------------------------------------
void RenderArea::setProjectionMode(const ProjectionMode proj)
//----------------------------------------------------------
{
  projectionMode_ = proj;
}

//----------------------------------------------------------
const RenderArea::ProjectionMode &RenderArea::getProjectionMode() const
//----------------------------------------------------------
{
  return projectionMode_;
}

//----------------------------------------------------------
void RenderArea::setRedrawTime(unsigned int t)
//----------------------------------------------------------
{
  redraw_time_ = t;
  if( timer()->isActive() ) timer()->start(redraw_time_);
}
  
//----------------------------------------------------------
const unsigned int &RenderArea::getRedrawTime() const 
//----------------------------------------------------------
{
  return redraw_time_;
}

//----------------------------------------------------------
void RenderArea::setIgnoreRedrawRequests(bool ignore)
//----------------------------------------------------------
{
  ignore_redraw_ = ignore;
}
      
//----------------------------------------------------------
const bool &RenderArea::ignoreRedrawRequests() const
//----------------------------------------------------------
{
  return ignore_redraw_;
}

//----------------------------------------------------------
void RenderArea::setProjection()
//----------------------------------------------------------
{
  if (projectionMode_ == PM_Ortho)
    setOrthoProjection();
  else
    setPerspectiveProjection();
}

//----------------------------------------------------------
void RenderArea::setPerspectiveProjection()
//----------------------------------------------------------
{
  //cerr << "Kbt::Osg::RenderArea::setPerspectiveProjection(): Called." << endl;
  //    fprintf(stderr, "Fovy: %f, z near: %f, z far: %f\n",
  //	    fovy_, z_near_, z_far_);
  float w = (vp_w_ > 0 ? (float)vp_w_ : (float)width());
  float h = (vp_h_ > 0 ? (float)vp_h_ : (float)height());
  float aspectRatio = w/h;
  osg::Matrix projection;
  projection.makePerspective(fovy_,aspectRatio,z_near_,z_far_);
    
  // Set to only part of frustum.
  //double left, right, bottom, top, z_near, z_far;
  //projection.getFrustum(left, right, bottom, top, z_near, z_far);
  //projection.makeFrustum(left, 0, 0, top, z_near, z_far);
  //projection.makeFrustum(0, right, 0, top, z_near, z_far);
  //projection.makeFrustum(left, 0, bottom, 0, z_near, z_far);
  //projection.makeFrustum(0, right, bottom, 0, z_near, z_far);
    
  getCamera()->setProjectionMatrix(projection);
}

//----------------------------------------------------------
void RenderArea::setOrthoProjection()
//----------------------------------------------------------
{
  //cerr << "Kbt::Osg::RenderArea::setOrthoProjection(): Called." << endl;
  float w = (vp_w_ > 0 ? (float)vp_w_ : (float)width());
  float h = (vp_h_ > 0 ? (float)vp_h_ : (float)height());
  float aspectRatio = w/h;
  osg::Matrix projection;

  projection.makeOrtho(-aspectRatio, aspectRatio, -1.0, 1.0, z_near_, z_far_);
  //projection.makeOrtho2D(-aspectRatio, aspectRatio, -1.0, 1.0); //, 0.0, 1.0);
        
  getCamera()->setProjectionMatrix(projection);
}
  
//----------------------------------------------------------
void RenderArea::setFieldOfView(double f)
//----------------------------------------------------------
{
  fovy_ = f;
  setProjection();
}
  
//----------------------------------------------------------
const double &RenderArea::getFieldOfView() const
//----------------------------------------------------------
{
  return fovy_;
}

//----------------------------------------------------------
void RenderArea::setZClippingPlane(double z_near, double z_far)
//----------------------------------------------------------
{
  z_near_ = z_near;
  z_far_  = z_far;
}

//----------------------------------------------------------
void RenderArea::getZClippingPlane(double &z_near, double &z_far) const
//----------------------------------------------------------
{
  z_near = z_near_;
  z_far  = z_far_;
}

//----------------------------------------------------------
int RenderArea::getViewportWidth() const
//----------------------------------------------------------
{
  return (vp_w_ > 0 ? vp_w_ : width());
}
  
//----------------------------------------------------------
int RenderArea::getViewportHeight() const
//----------------------------------------------------------
{
  return (vp_h_ > 0 ? vp_h_ : height());
}
  
//----------------------------------------------------------
QImage RenderArea::grabImage()
//----------------------------------------------------------
{
  QGLWidget::makeCurrent();
    
  QImage res;
    
  if ( !format().rgba() )
    return res;
    
  int w = width();
  int h = height();
  int x = 0;
  int y = 0;
  if(vp_w_ > 0 && vp_h_ > 0)
    {
      w = vp_w_;
      h = vp_h_;
      x = (width() - w)/2;
      y = (height() - h)/2;
    }
    
  res = QImage( w, h, QImage::Format_RGB32 );
  glReadPixels( x, y, w, h, GL_RGBA, GL_UNSIGNED_BYTE, res.bits() );
    
  if ( QSysInfo::ByteOrder == QSysInfo::BigEndian )
  {
    // OpenGL gives RGBA; Qt wants ARGB
    uint *p = (uint*)res.bits();
    uint *end = p + w*h;
    while ( p < end )
      *p++ >>= 8;
  }
  else
  {
    // OpenGL gives ABGR (i.e. RGBA backwards); Qt wants ARGB
    res = res.rgbSwapped();
  }
    
  return res.mirrored();
}
  
//----------------------------------------------------------
void RenderArea::setViewportSize(int w, int h)
//----------------------------------------------------------
{
  vp_w_ = w;
  vp_h_ = h;
    
  // Trigger a recalc
  resizeGL(width(), height()); 
}

//----------------------------------------------------------
void RenderArea::setClearColor(const osg::Vec4& color) 
//----------------------------------------------------------
{
  _camera->setClearColor(color); 
}
  
//----------------------------------------------------------
QTimer *RenderArea::timer()
//----------------------------------------------------------
{
  if(!qtimer_)
    {
      qtimer_=new QTimer(this);
      connect(qtimer_,SIGNAL(timeout()),this,SLOT(doIdleAnimationFrame()));
    }
  return qtimer_;
}
  
//----------------------------------------------------------
void RenderArea::forceUpdateGL()
//----------------------------------------------------------
{
  AdapterWidget::updateGL();
}
  
//----------------------------------------------------------
void RenderArea::updateGL()
//----------------------------------------------------------
{
  if(ignore_redraw_ || redrawPolicy_==CONSTANT || needContinuousUpdate_) return; 
    
  AdapterWidget::updateGL();
}
  
//----------------------------------------------------------
void RenderArea::requestRedraw()
//----------------------------------------------------------
{
  // If continuous mode is on redraw will happen soon. No need to do anything.
  if(redrawPolicy_==CONSTANT || needContinuousUpdate_) return; 
    
  if(!needRedraw_) 
    {
      // First time requestRedraw has been called since start of last paintGL.
      // Schedule an updateGL call.
      needRedraw_ = true; 
      QTimer::singleShot(0, this,SLOT(doIdleAnimationFrame()));
    }
}
    
//----------------------------------------------------------
void RenderArea::requestContinuousUpdate(bool needed)
//----------------------------------------------------------
{
  needContinuousUpdate_ = needed;
    
  if(redrawPolicy_==AS_NEEDED)
    {
      if(needed)
	{
	  if( !timer()->isActive() ) timer()->start(redraw_time_);
	}
      else
	{
	  if( timer()->isActive() ) timer()->stop();
	}
    }
}

//----------------------------------------------------------
const bool &RenderArea::getNeedContinuousUpdate() const
//----------------------------------------------------------
{
  return needContinuousUpdate_;
}
  
//----------------------------------------------------------
void RenderArea::doIdleAnimationFrame()
//----------------------------------------------------------
{
  forceUpdateGL();
}

//----------------------------------------------------------
void RenderArea::removeEventHandler(osgGA::GUIEventHandler *eventHandler)
//----------------------------------------------------------
{
  for (EventHandlers::iterator it = _eventHandlers.begin(); it != _eventHandlers.end(); ++it) {
    osgGA::GUIEventHandler *ev = (*it).get();
    if (ev == eventHandler) {
      _eventHandlers.erase(it);
      break;
    }
  }
}

//----------------------------------------------------------
bool RenderArea::hasEventHandler(osgGA::GUIEventHandler *eventHandler)
//----------------------------------------------------------
{
  for (EventHandlers::iterator it = _eventHandlers.begin(); it != _eventHandlers.end(); ++it) {
    osgGA::GUIEventHandler *ev = (*it).get();
    if (ev == eventHandler) {
      return true;
    }
  }
  return false;
}

//----------------------------------------------------------
void RenderArea::setCurrObject(osg::MatrixTransform* obj)
//----------------------------------------------------------
{
  curr_obj_ = obj; 
   
  saveObjectAndCamera(obj);
}

//----------------------------------------------------------
static osg::Matrix getCombinedTransform(osg::MatrixTransform* tnode)
//----------------------------------------------------------
{
  osg::Matrix mat_in = tnode->getMatrix();
    
  osg::Node::ParentList plist = tnode->getParents();
  osg::Node::ParentList::iterator it;

  osg::Matrix mat = osg::Matrix::identity();

  for (it=plist.begin(); it != plist.end(); it++) {
    osg::MatrixTransform* tnode = dynamic_cast<osg::MatrixTransform*>(*it);
      
    if (tnode)
      mat.preMult(tnode->getMatrix());
  }
    
  mat.preMult(mat_in);
    
  return mat;
}

//----------------------------------------------------------
void RenderArea::setTrackballCenter(osg::MatrixTransform* obj)
//----------------------------------------------------------
{
  //osg::Matrixd mat = obj->getMatrix();
  osg::Matrixd mat = getCombinedTransform(obj);

  TrackballManipulator *trackball = getTrackballManipulator();

  orig_obj_mat_combined_.postMult(trackball->getInverseMatrix());
    
  osg::Vec3 center = mat.preMult(obj->computeBound().center());
    
  //cerr << "RenderArea::setTrackballCenter(): center = " 
  //	   << center.x() << ", " << center.y() << ", " << center.z() << endl;
    
  trackball->setCenter(center);
    
  osg::Vec3 eye, up;
  trackball->getMatrix().getLookAt(eye, center, up);
  //      printf("tb->center = %lf %lf %lf...\n", center.x(), center.y(), center.z());
    
  orig_obj_mat_combined_.postMult(trackball->getMatrix());
}

//----------------------------------------------------------
void RenderArea::saveObjectAndCamera(osg::MatrixTransform* tnode)
//----------------------------------------------------------
{
  if (!tnode) return;

  orig_obj_mat_          = tnode->getMatrix();
  orig_obj_mat_combined_ = getCombinedTransform(tnode);

  TrackballManipulator *trackball = getTrackballManipulator();
    
  orig_proj_mat_  = trackball->getMatrix();
    
  osg::Vec3 eye, center, up, obj_center;
  //      osg::Matrixd::inverse(orig_proj_mat_).getLookAt(eye, center, up);
  orig_proj_mat_.getLookAt(eye, center, up);
    
  obj_center = orig_obj_mat_combined_.preMult(tnode->computeBound().center());
  //   obj_center = tnode->getBoundingBox().center();
    
  orig_proj_mat_  = trackball->getMatrix();
}

