#ifndef OsgQt_RenderArea_h_Included             // -*- C++ -*-
#define OsgQt_RenderArea_h_Included   

#include <osgQt/AdapterWidget.h>

namespace osg {
  class MatrixTransform;
}

class TrackballManipulator;

/// RenderArea -- QT viewer. Based on Osg example program.
/**
 * QT viewer. Based on Osg example program.
 *
 * This class inherits from both Qt and Open Scene Graph. It makes shure
 * the openGL commands from osg are executed correctly by calling 
 * osgViewer::ViewerBase::frame() from within QGLWidget::paintGL()
 *
 * \ingroup Osg
 *
 */

class RenderArea : public AdapterWidget, public osgViewer::Viewer
{
  Q_OBJECT
  
public:
  
  RenderArea(QWidget * parent = 0, const char * name = 0, 
             const QGLWidget * shareWidget = 0, WindowFlags f = 0);
  
  virtual ~RenderArea();
  
  void projectWindowXYIntoObject(int x, int y, osg::Vec3& near_point, osg::Vec3& far_point) const;
  
  /** The RedrawPolicy controls how the window is updated. */
  enum RedrawPolicy 
  {
    AS_NEEDED,      /** Only redraw as needed (Default), e.g. when a window is popped forward, or model is 'spun'. */
    CONSTANT        /** Redraw constantly, a la <i>osgviewer</i>. (Good for models with animation.) */
  };

  
  /** The ProjectionMode controls how the camera renders thw view. */
  enum ProjectionMode 
  {
    PM_Perspective, /** Perspective projection */
    PM_Ortho        /** Othographic projection */
  };
  
  /** Set's the redraw mode on the viewer */
  void setRedrawPolicy(const RedrawPolicy pol);
  
  /** Get the current redraw mode */
  RedrawPolicy getRedrawPolicy() const;
  
  /** Set the projection mode */
  void setProjectionMode(const ProjectionMode proj);
  
  /** Returns the projection mode */
  const ProjectionMode &getProjectionMode() const;
  
  /** Set approx time between frames in milliseconds. */
  void setRedrawTime(unsigned int t);
      
  /** Set approx time between frames in milliseconds. */
  const unsigned int &getRedrawTime() const;
      
  /** Set if redraw requests should be ignored. This is done request level for redraw updates. */
  void setIgnoreRedrawRequests(bool ignore);
      
  /** Get flag indicating if redraw requests should be ignored.*/
  const bool &ignoreRedrawRequests() const;
      
  void setFieldOfView(double f);
      
  const double &getFieldOfView() const;

  void setZClippingPlane(double z_near, double z_far);

  void getZClippingPlane(double &z_near, double &z_far) const;

//  osgGA::MatrixManipulator* getMatrixManipulator();
      
//  void setMatrixManipulator(osgGA::MatrixManipulator* manip);
      
  TrackballManipulator* getTrackballManipulator();
      
  int getViewportWidth() const;
      
  int getViewportHeight() const;
      
  QImage grabImage();
      
  void setViewportSize(int w, int h);
      
  void setClearColor(const osg::Vec4& color);
            
  void setCurrObject(osg::MatrixTransform* obj);
      
  void forceUpdateGL();
      
  virtual void updateGL();
      
  virtual void requestRedraw();
    
  virtual void requestContinuousUpdate(bool needed=true);
    
  const bool &getNeedContinuousUpdate() const;

  void setTrackballCenter(osg::MatrixTransform* obj);
    
  void removeEventHandler(osgGA::GUIEventHandler *eventHandler);

  bool hasEventHandler(osgGA::GUIEventHandler *eventHandler);
    
signals:
    
  void viewerResized(QGLWidget* w, int, int);
    
  void enteringPaintGL();
  void leavingPaintGL();
			 
protected slots:
    
  void doIdleAnimationFrame();
    
protected:
    
  virtual void processAdaptedEvents();
    
  void updateTimer();
    
  /*
   * QtGL methods
   */
  virtual void initializeGL();
  virtual void resizeGL(int,int);
  virtual void paintGL();
    
  /**
   * Virtual method used by paintGL to set the correct model view
   * for each frame. Default version uses the scenegraphs matrix
   * manipulator. 
   */
  virtual void updateModelView();

  void setProjection();
  void setPerspectiveProjection();
  void setOrthoProjection();
    
  /** Computes the matrix product ModelView * Projecion * WindowMatrix
   *  which converts from local_coordinates to window coordinates
   *  See the documentation of osg::ViewPort::computeWindowMatrix
   */
  const osg::Matrix computeMVPW() const;
    
  void saveObjectAndCamera(osg::MatrixTransform* tnode);
    
  double fovy_;
    
  // We need a high res timer so we can update the frame stamp on
  // every frame. This is not related to qtimer, which is used to
  // set up a callback to handle frame updates.
  // Setting up a timer in a class should be easier than this:
  static osg::Timer timer_;
  static const osg::Timer_t initialTick_;
    
  QTimer *timer();    // Ensures that qtimer_ exists
  QTimer *qtimer_;	// This  should always be accessed via the timer() method
    
//  osg::ref_ptr<osgGA::MatrixManipulator> matrixManipulator_;
    
  osg::ref_ptr<osg::MatrixTransform> curr_obj_;
    
  /** Actual viewport size, (may be smaller than widget) */
  int vp_w_, vp_h_;
    
  RedrawPolicy redrawPolicy_;
  ProjectionMode projectionMode_;
  unsigned int redraw_time_;
    
  bool needContinuousUpdate_;
  bool needRedraw_;
  bool ignore_redraw_;
    
  /** Near and far clipping planes */
  double z_near_, z_far_;
    
  osg::Matrixd orig_obj_mat_;
  osg::Matrixd orig_obj_mat_combined_;
  osg::Matrixd orig_proj_mat_;
};

#endif // OsgQt_RenderArea_h_Included
