#include <scenegraph/OsgDolfinObject.h>

#include <osg/Geode>
#include <osg/Geometry>

#include <osg/PrimitiveSet>
#include <osg/Material>

#include <osg/Group>
#include <osg/Notify>
#include <osg/Material>
#include <osg/PolygonOffset>
#include <osg/PolygonMode>
#include <osg/LineStipple>
#include <osg/LineWidth>
#include <osg/Shape>
#include <osg/ShapeDrawable>
#include <osg/Point>

#include <osg/BoundingSphere>
#include <osg/CullFace>

#include <osgUtil/Optimizer>
#include <osgUtil/SmoothingVisitor>

#include <iostream>
#include <dolfin/mesh/Vertex.h>
#include <dolfin/mesh/Face.h>
#include <dolfin/mesh/Edge.h>

#include <QDebug>

//----------------------------------------------------------
OsgDolfinObject::OsgDolfinObject(DolfinObject& dolfin_object)
  : vertex_size_(3), edge_size_(2), draw_vertices_(true), 
    draw_edges_(true), dolfin_object(dolfin_object)
//----------------------------------------------------------
{
  setName("OsgDolfinObject");
}
//----------------------------------------------------------
OsgDolfinObject::OsgDolfinObject(const OsgDolfinObject& obj)
  : OsgMeshObject(obj),
    vertex_size_(obj.vertex_size_), 
    edge_size_(obj.edge_size_),
    draw_vertices_(obj.draw_vertices_), 
    draw_edges_(obj.draw_edges_),
    dolfin_object(obj.dolfin_object)
//----------------------------------------------------------
{
  setName("OsgDolfinObject");
}

//----------------------------------------------------------
OsgDolfinObject::~OsgDolfinObject()
//----------------------------------------------------------
{

}
//----------------------------------------------------------
void OsgDolfinObject::initialize()
//----------------------------------------------------------
{
  OsgMeshObject::initialize();

  vertex_size_ = 3;
  edge_size_   = 2;

  //=================================
  // Create color palette
  //=================================

  color_palette_ = new osg::Vec4Array();
  color_palette_->push_back(osg::Vec4(0.4, 0.4, 0.4, 1.0)); // default gray
  color_palette_->push_back(osg::Vec4(1.0, 0.0, 0.0, 1.0)); // 1 = red
  color_palette_->push_back(osg::Vec4(0.0, 1.0, 0.0, 1.0)); // 2 = green
  color_palette_->push_back(osg::Vec4(0.0, 0.0, 1.0, 1.0)); // 3 = blue
  color_palette_->push_back(osg::Vec4(1.0, 1.0, 0.0, 1.0)); // 4 = yellow
  color_palette_->push_back(osg::Vec4(0.0, 1.1, 1.1, 1.0)); // 5 = cyan
  color_palette_->push_back(osg::Vec4(1.0, 0.0, 1.0, 1.0)); // 6 = purple
  color_palette_->push_back(osg::Vec4(1.0, 1.0, 1.0, 1.0)); // 7 = white
  color_palette_->push_back(osg::Vec4(0.0, 0.0, 0.0, 1.0)); // 8 = black

  //=================================
  // Read vertices
  //=================================

  std::cout << "DolfinObject::initialize: Read vertices" << std::endl;
  readVertices();
  std::cout << "DolfinObject::initialize: Read indices" << std::endl;
  readIndices();

  //=================================
  // Read indices
  //=================================

  std::cout << "OsgDolfinObject::initialize: Build OSG Node" << std::endl;
  buildOSGNode();
  setDataOk(true);

  //=================================
  // Finalize geometry
  //=================================

  std::cout << "OsgDolfinObject::initialize: Finalize geometry" << std::endl;
  OsgMeshObject::finalize();

  //=================================
  // Update parameters, colors, etc.
  //=================================

  std::cout << "OsgDolfinObject::initialize: Update color and intensity" << std::endl;
  updateColorAndIntensity();

  std::cout << "OsgDolfinObject::initialize: Update transparency" << std::endl;
  updateTransparency();

  std::cout << "OsgDolfinObject::initialize: Update cull mode" << std::endl;
  updateCullMode();

  std::cout << "OsgDolfinObject::initialize: Update twosided ligthing" << std::endl;
  updateTwoSidedLighting();

  std::cout << "OsgDolfinObject::initialize: Update wireframe color" << std::endl;
  updateWireframeColor();

  std::cout << "OsgDolfinObject::initialize: Update boundary indicator colors" << std::endl;
  updateBoundaryIndicatorColors();

  std::cout << "OsgDolfinObject::initialize: Finalize transform node" << std::endl;
  OsgTransformObject::finalize();

  updateHighlight();

  std::cout << "OsgDolfinObject::initialize: Done." << std::endl;
}

//----------------------------------------------------------
bool OsgDolfinObject::readVertices()
//----------------------------------------------------------
{
  std::cout << "OsgDolfinObject::readVertices" << std::endl;

  // Copy vertex to osg
  double x = 0.0;
  double y = 0.0;
  double z = 0.0;

  vertices_ = new osg::Vec3Array();

  for (dolfin::VertexIterator v(dolfin_object.surface_mesh()); !v.end(); ++v) {
    x = v->point().x();
    y = v->point().y();
    z = v->point().z();
    vertices_->push_back(osg::Vec3(x,y,z));
 }

 return true;
}

//----------------------------------------------------------
static osg::Vec3 computeNormal(const dolfin::Face &f)
//----------------------------------------------------------
{
  const dolfin::uint* entities = f.entities(0);
  dolfin::uint  num_entities = f.num_entities(0);

  osg::Vec3 n(0.0, 0.0, 0.0);

  if (num_entities>=3) 
  {
    dolfin::Vertex v1(f.mesh(), entities[0]);
    dolfin::Vertex v2(f.mesh(), entities[1]);
    dolfin::Vertex v3(f.mesh(), entities[2]);

    osg::Vec3 p1(v1.point().x(), v1.point().y(), v1.point().z());
    osg::Vec3 p2(v2.point().x(), v2.point().y(), v2.point().z());
    osg::Vec3 p3(v3.point().x(), v3.point().y(), v3.point().z());

    n = (p3 - p2) ^ (p1 - p2);
    n.normalize();
  }

  return n;
}

//================================================================================

//----------------------------------------------------------
bool OsgDolfinObject::readIndices()
//----------------------------------------------------------
{
  std::cout << "OsgDolfinObject::readIndices" << std::endl;

  //=====================================
  // Set up indices for faces (polygons)
  //=====================================

  // NOTE
  // When quad-based meshes are introduced in Dolfin,
  // this might need a rewrite. There is currently no mesh available
  // to test on.
  //
  // If there are plans to create meshes with a VARIABLE number of vertices per face
  // then a rewrite is really necessary. A plan to make this work would be to create a
  // osg primitive set (as below) for each category of polygons (triangles, quads, whatnot)
  // and draw all of these sets. That would probably work out ok.
  
  dolfin::Face face(dolfin_object.surface_mesh(), 0);
  uint num_entities = face.num_entities(0);
  
  if(num_entities == 3) 
  {
    osg_dolfin_faces_ = new osg::DrawElementsUInt(osg::PrimitiveSet::TRIANGLES, 0);
  }
  else if(num_entities == 4) 
  {
    osg_dolfin_faces_= new osg::DrawElementsUInt(osg::PrimitiveSet::QUADS, 0);
    std::cout << "OsgDolfinObject::ReadIndices: Warning: This mesh has 4 vertices per face which is not properly tested. " << std::endl;
  } else 
  {
    std::cout << "OsgDolfinObject::ReadIndices: Unsupported mesh with " << num_entities << " vertices per face " << std::endl;
    return false;
  }

  face_colors_ = new osg::Vec4Array();

  for (dolfin::FaceIterator f(dolfin_object.surface_mesh()); !f.end(); ++f) 
  {
    const uint* entities = f->entities(0);
    uint  num_entities = f->num_entities(0);
    //uint  index = f->index();
    
    face_colors_->push_back(osg::Vec4(0.4,0.4,0.4,1.0));
    
    if(num_entities == 3) {
      osg_dolfin_faces_->push_back((unsigned int) entities[0]);
      osg_dolfin_faces_->push_back((unsigned int) entities[1]);
      osg_dolfin_faces_->push_back((unsigned int) entities[2]);
    }
    else if(num_entities == 4) {
      osg_dolfin_faces_->push_back((unsigned int) entities[0]);
      osg_dolfin_faces_->push_back((unsigned int) entities[1]);
      osg_dolfin_faces_->push_back((unsigned int) entities[2]);
      osg_dolfin_faces_->push_back((unsigned int) entities[3]);
    }
    else {
      std::cerr << "DolfinObject: Trouble, this is not a triangular or quad face, num_entities = " << num_entities << std::endl;
      return false;
    }
  }

  //================
  // set up normals
  //================

  {
    normals_ = new osg::Vec3Array();
    osg::Vec3 n;
    int count;

    //Note: It seems that the orientation of the faces in a BoundaryMesh is not
    //      constistent, some faces are oriented forwards and some faces are oriented
    //      backwards. This is why the smoothing is not working properly.
#ifdef NORMALS_PER_VERTEX
    for (dolfin::VertexIterator v(dolfin_object.surface_mesh()); !v.end(); ++v) {
      dolfin::Vertex vi = *v;
      n.set(0.0, 0.0, 0.0);
      count = 0;
      for (dolfin::FaceIterator f(vi); !f.end(); ++f, ++count) 
      {
        n += computeNormal(*f);
      }
      if (count==0)
	continue;
      n /= (double)count;
      n.normalize();
      normals_->push_back(n);
      std::cerr << "Normal #" << vi.index() << " = " << n[0] << " " << n[1] << " " << n[2] << " l=" << n.length() << std::endl;
    }
#else
    count = 0;
    for (dolfin::FaceIterator f(dolfin_object.surface_mesh()); !f.end(); ++f, ++count) 
    {
      n = computeNormal(*f);
      normals_->push_back(n);
    }
#endif
  }

  //==============
  // set up edges
  //==============

  std::cout << "OsgDolfinObject::readIndices: Surface mesh has " << dolfin_object.surface_mesh().size(1) << " edges. " << std::endl;
  
  edge_colors_ = new osg::Vec4Array();
  
  osg_dolfin_edges_ = new osg::DrawElementsUInt(osg::PrimitiveSet::LINES, 0);
  for (dolfin::EdgeIterator f(dolfin_object.surface_mesh()); !f.end(); ++f) 
  {
    const uint* entities = f->entities(0);
    uint  num_entities = f->num_entities(0);
    //uint  index = f->index();
    
    edge_colors_->push_back(osg::Vec4(0.4,0.4,0.4,1.0));
    
    if(num_entities == 2) 
    {
      osg_dolfin_edges_->push_back((unsigned int) entities[0]);
      osg_dolfin_edges_->push_back((unsigned int) entities[1]);
    }
    else 
    {
      std::cout << "DolfinObject: Trouble, the lines don't have exactly two points.. ???" << std::endl;
      return false;
    }
  }

  //==========================
  // set up points (vertices)
  //==========================

  std::cout << "OsgDolfinObject::readIndices: Surface mesh has " << dolfin_object.surface_mesh().size(0) << " vertices. " << std::endl;
  
  vertex_colors_ = new osg::Vec4Array();
  
  osg_dolfin_vertices_ = new osg::DrawElementsUInt(osg::PrimitiveSet::POINTS, 0);
  for (dolfin::VertexIterator f(dolfin_object.surface_mesh()); !f.end(); ++f) 
  {
    //const uint* entities = f->entities(0);
    //uint  num_entities = f->num_entities(0);
    uint  index = f->index();
    
    vertex_colors_->push_back(osg::Vec4(0.4,0.4,0.4,1.0));
    
    osg_dolfin_vertices_->push_back((unsigned int) index);
  }

  return true;
}

//================================================================================

//----------------------------------------------------------
bool OsgDolfinObject::buildOSGNode()
//----------------------------------------------------------
{
  std::cerr << "OsgDolfinObject::buildOSGNode" << std::endl;
  std::cout << "OsgDolfinObject::buildOSGNode: # vertex array   = " << vertices_->size() << std::endl;
  std::cout << "OsgDolfinObject::buildOSGNode: # vertex indices = " << osg_dolfin_vertices_->size() << std::endl;
  std::cout << "OsgDolfinObject::buildOSGNode: # edge indices   = " << osg_dolfin_edges_->size()/2 << std::endl; // because there are two vertices per edge
  std::cout << "OsgDolfinObject::buildOSGNode: # face indices   = " << osg_dolfin_faces_->size()/3 << std::endl; // because there are three vertices per face

  //===========
  // Add faces
  //===========
  
  geode_ = new osg::Geode();
  mesh_ = new osg::Geometry();
  
  geode_->addDrawable(mesh_.get());
  mesh_->setVertexArray(vertices_.get());
  mesh_->setNormalArray(normals_.get());
#ifdef NORMALS_PER_VERTEX
  mesh_->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);
#else
  mesh_->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);
#endif
  mesh_->addPrimitiveSet(osg_dolfin_faces_.get());
  mesh_->setColorArray(face_colors_.get());
  mesh_->setColorBinding(osg::Geometry::BIND_PER_PRIMITIVE);
  
  //osgUtil::SmoothingVisitor sv;
  //sv.smooth(*mesh_);
  
  addChild(geode_.get());
  
  //===========
  // Add edges
  //===========
  
  edgeGeode_ = new osg::Geode();
  edge_      = new osg::Geometry();

  edgeGeode_->addDrawable(edge_.get());
  edge_->setVertexArray(vertices_.get());
  edge_->addPrimitiveSet(osg_dolfin_edges_.get());
  edge_->setColorArray(edge_colors_.get());
  edge_->setColorBinding(osg::Geometry::BIND_PER_PRIMITIVE);
  
  addChild(edgeGeode_.get());

  //=======================
  // Add vertices (points)
  //=======================

  vertexGeode_ = new osg::Geode();
  vertex_      = new osg::Geometry();

  vertexGeode_->addDrawable(vertex_.get());
  vertex_->setVertexArray(vertices_.get());
  vertex_->addPrimitiveSet(osg_dolfin_vertices_.get());
  vertex_->setColorArray(vertex_colors_.get());
  vertex_->setColorBinding(osg::Geometry::BIND_PER_VERTEX);

  addChild(vertexGeode_.get());

  return true;
}

//================================================================================

//----------------------------------------------------------
void OsgDolfinObject::updateTransparency()
//----------------------------------------------------------
{
  osg::Vec4Array* vec = face_colors_.get();

  if(vec)
  {
    const SimResState &current_state = getCurrentState();
    double transparency = current_state.getTransparency();
    
    osg::StateSet* stateset = geode_->getOrCreateStateSet();
    
    // add rendering hint
    if(transparency != 0.0)
      stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
    else
      stateset->setRenderingHint(osg::StateSet::OPAQUE_BIN);
    
    osg::Vec4Array::iterator it;
    for(it=vec->begin() ; it != vec->end() ; it++)
      it->w() = 1.0-transparency;

    mesh_->setColorArray(vec);
    
    ambient().w()  = 1.0-transparency;
    diffuse().w()  = 1.0-transparency;
    specular().w() = 1.0-transparency;
    emission().w() = 1.0-transparency;

    updateMaterial();
  }
}

//================================================================================

//----------------------------------------------------------
void OsgDolfinObject::updateMaterial()
//----------------------------------------------------------
{
  //std::cout << "OsgDolfinObject::updateMaterial" << std::endl;

  if(!isDataOk()) return;

  // Update base color
  
  const SimResState &current_state = getCurrentState();
  
  double transparency = 1.0 - current_state.getTransparency();
  (*color_palette_)[0] = osg::Vec4(current_state.getBaseColor().red()/255.0,
				   current_state.getBaseColor().green()/255.0,
				   current_state.getBaseColor().blue()/255.0,
				   transparency);
  

  //===============
  // Face material
  //===============
  
  double shininess = current_state.getShininess();
  
  material_->setAmbient(osg::Material::FRONT_AND_BACK, ambient());
  material_->setDiffuse(osg::Material::FRONT_AND_BACK, diffuse());
  material_->setSpecular(osg::Material::FRONT_AND_BACK, specular());
  material_->setEmission(osg::Material::FRONT_AND_BACK, emission());
  material_->setShininess(osg::Material::FRONT_AND_BACK, shininess);
  
  osg::StateSet* stateset = NULL;
  if(mesh_.get())
  {
    //        std::cout << "OsgDolfinObject::updateMaterial: stateset for mesh" << std::endl;
    stateset = mesh_->getOrCreateStateSet();
  }
  else
    stateset = geode_->getOrCreateStateSet();

  material_->setColorMode(osg::Material::AMBIENT_AND_DIFFUSE);
  stateset->setAttributeAndModes(material_.get(),osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON);
  stateset->setMode(GL_LIGHTING,osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON);
  
  //===============
  // Edge material
  //===============
  
  stateset = NULL;
  if(edge_.get())
  {
    //        std::cout << "OsgDolfinObject::updateMaterial: stateset for edge" << std::endl;
    stateset = edge_->getOrCreateStateSet();
  }
  else
    stateset = edgeGeode_->getOrCreateStateSet();
  
  osg::PolygonOffset *edgeoffset = new osg::PolygonOffset();
  edgeoffset->setFactor(-1.0);
  edgeoffset->setUnits(-2.0);
  stateset->setAttribute(edgeoffset);
  
  osg::LineWidth *lw = new osg::LineWidth(edge_size_);
  stateset->setAttribute(lw);
  stateset->setMode(GL_LIGHTING,osg::StateAttribute::OVERRIDE|osg::StateAttribute::OFF);
  
  //===============
  // Vertex material
  //===============
  
  stateset = NULL;
  if(vertex_.get())
  {
    //        std::cout << "OsgDolfinObject::updateMaterial: stateset for vertex" << std::endl;
    stateset = vertex_->getOrCreateStateSet();
  }
  else
    stateset = vertexGeode_->getOrCreateStateSet();
  
  osg::Point *point = new osg::Point();
  point->setSize(vertex_size_);
  stateset->setAttribute(point);
  
  osg::PolygonOffset *offset = new osg::PolygonOffset();
  offset->setFactor(-1.0);
  offset->setUnits(-2.0);
  stateset->setAttribute(offset);
  
  stateset->setMode(GL_POINT_SMOOTH, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON);
  stateset->setMode(GL_LIGHTING,     osg::StateAttribute::OVERRIDE|osg::StateAttribute::OFF);
}

//================================================================================

//============================================================
// Vertex Boundary indicator functions
//============================================================

//----------------------------------------------------------
void OsgDolfinObject::setVertexBoundaryIndicator(dolfin::uint id, int value)
//----------------------------------------------------------
{
  // Update the visualization
  // Set color on the osg mesh
  (*vertex_colors_)[id] = (*color_palette_)[std::min(value,8)];
  vertex_->setColorArray(vertex_colors_.get());
}
/*
//----------------------------------------------------------
void OsgDolfinObject::fillVertexBoundaryIndicator(dolfin::uint id, int value)
//----------------------------------------------------------
{
  recursiveVertexFill(id, value);
  updateBoundaryIndicatorColors();
}
*/
/*
//----------------------------------------------------------
void OsgDolfinObject::fillVertexSelectedIndicator(dolfin::uint id, int value)
//----------------------------------------------------------
{
  dolfin::uint dolfin_index = dolfin_vertices_->values()[id];
  int hitvalue = vertex_boundary_indicators_.values()[dolfin_index];
  recursiveVertexFillSelected(id, hitvalue, value);
  updateBoundaryIndicatorColors();
}

//----------------------------------------------------------
void OsgDolfinObject::recursiveVertexFillSelected(dolfin::uint id, int hitvalue, int value)
//----------------------------------------------------------
{
  dolfin::uint dolfin_index = dolfin_vertices_->values()[id];
  int indicator = vertex_boundary_indicators_.values()[dolfin_index];
  
  if((indicator != value) && (indicator == hitvalue))
    {
      vertex_boundary_indicators_[dolfin_index] = value;
      dolfin::MeshEntity vertex(dolfin_object.surface_mesh(), 0, id);
      for(dolfin::MeshEntityIterator v(vertex, 0); !v.end(); ++v)
	recursiveVertexFillSelected(v->index(), hitvalue, value);
    }
}
*/
// //----------------------------------------------------------
// void OsgDolfinObject::setVertexAllBoundaryIndicators(int value)
// //----------------------------------------------------------
// {
//   for(dolfin::VertexIterator v(dolfin_object.surface_mesh()); !v.end(); ++v)
//   {
//     uint id = v->index();

//     // Update the mesh function
//     dolfin_object.setVertexBoundaryIndicator(id, value);

//     // Update the visualization
//     setVertexBoundaryIndicator(id, value);
//   }
//   updateBoundaryIndicatorColors();
// }


//============================================================
// Edge Boundary indicator functions
//============================================================

//----------------------------------------------------------
void OsgDolfinObject::setEdgeBoundaryIndicator(uint id, int value)
//----------------------------------------------------------
{
  // Set color on the osg mesh
  (*edge_colors_)[id] = (*color_palette_)[std::min(value,8)];
  edge_->setColorArray(edge_colors_.get());
}

// //----------------------------------------------------------
// void OsgDolfinObject::fillEdgeBoundaryIndicator(uint id, int value)
// //----------------------------------------------------------
// {
//   recursiveEdgeFill(id, value);
//   updateBoundaryIndicatorColors();
// }
/*
//----------------------------------------------------------
void OsgDolfinObject::fillEdgeSelectedIndicator(dolfin::uint id, int value)
//----------------------------------------------------------
{
  dolfin::uint dolfin_index = dolfin_edges_.values()[id];
  int hitvalue = edge_boundary_indicators_.values()[dolfin_index];
  recursiveEdgeFillSelected(id, hitvalue, value);
  updateBoundaryIndicatorColors();
}

//----------------------------------------------------------
void OsgDolfinObject::recursiveEdgeFillSelected(dolfin::uint id, int hitvalue, int value)
//----------------------------------------------------------
{
  dolfin::uint dolfin_index = dolfin_edges_.values()[id];
  int indicator = edge_boundary_indicators_.values()[dolfin_index];
  
  if((indicator != value) && (indicator == hitvalue))
  {
    edge_boundary_indicators_[dolfin_index] = value;
    dolfin::MeshEntity edge(dolfin_object.surface_mesh(), 1, id);
    for(dolfin::MeshEntityIterator v(edge, 1); !v.end(); ++v)
      recursiveEdgeFillSelected(v->index(), hitvalue, value);
  }
}
*/
// //----------------------------------------------------------
// void OsgDolfinObject::setEdgeAllBoundaryIndicators(int value)
// //----------------------------------------------------------
// {
//   for(dolfin::EdgeIterator v(dolfin_object.surface_mesh()); !v.end(); ++v)
//   {
//     uint id = v->index();
//     // Update the mesh function and the visualization
//     setEdgeBoundaryIndicator(id, value);
//   }
//   updateBoundaryIndicatorColors();
// }



//============================================================
// Face Boundary indicator functions
//============================================================

//----------------------------------------------------------
void OsgDolfinObject::setFaceBoundaryIndicator(uint id, int value)
//----------------------------------------------------------
{   
  (*face_colors_)[id] = (*color_palette_)[std::min(value,8)];
  (*face_colors_)[id].w() = 1.0 - getCurrentState().getTransparency();
  mesh_->setColorArray(face_colors_.get());
}

// NOTE: This should be removed.

// //----------------------------------------------------------
// void OsgDolfinObject::recursiveFaceFill(uint id, int value)
// //----------------------------------------------------------
// {
//   int indicator = dolfin_object.getFaceBoundaryIndicator(id); 

//   if(indicator != value)
//   {
//     // Update mesh function and visualization
//     setFaceBoundaryIndicator(id, value);

//     dolfin::MeshEntity face(dolfin_object.surface_mesh(), 2, id);
//     for(dolfin::MeshEntityIterator v(face, 2); !v.end(); ++v)
//       recursiveFaceFill(v->index(), value);
//   }
// }
// //----------------------------------------------------------
// void OsgDolfinObject::fillFaceBoundaryIndicator(uint id, int value)
// //----------------------------------------------------------
// {
//   recursiveFaceFill(id, value);
//   updateBoundaryIndicatorColors();
// }

/*
//----------------------------------------------------------
void OsgDolfinObject::fillFaceSelectedIndicator(dolfin::uint id, int value)
//----------------------------------------------------------
{
  dolfin::uint dolfin_index = dolfin_faces_->values()[id];
  int hitvalue = face_boundary_indicators_.values()[dolfin_index];
  recursiveFaceFillSelected(id, hitvalue, value);
  updateBoundaryIndicatorColors();
}

//----------------------------------------------------------
void OsgDolfinObject::recursiveFaceFillSelected(dolfin::uint id, int hitvalue, int value)
//----------------------------------------------------------
{
  dolfin::uint dolfin_index = dolfin_faces_->values()[id];
  int indicator = face_boundary_indicators_.values()[dolfin_index];
  
  if((indicator != value) && (indicator == hitvalue))
  {
    face_boundary_indicators_[dolfin_index] = value;
    dolfin::MeshEntity face(dolfin_object.surface_mesh(), 2, id);
    for(dolfin::MeshEntityIterator v(face, 2); !v.end(); ++v)
      recursiveFaceFillSelected(v->index(), hitvalue, value);
  }
}
*/
//============================================================
// Boundary indicator color setting
//============================================================

//----------------------------------------------------------
void OsgDolfinObject::drawVertices(bool set)
//----------------------------------------------------------
{
  if(vertexGeode_.get())
  {
    if(set)
      vertexGeode_->setNodeMask(0xffffffff);
    else
      vertexGeode_->setNodeMask(0);
  }
}

//----------------------------------------------------------
void OsgDolfinObject::drawEdges(bool set)
//----------------------------------------------------------
{
  if(edgeGeode_.get())
  {
    if(set)
      edgeGeode_->setNodeMask(0xffffffff);
    else
      edgeGeode_->setNodeMask(0);
  }
}

//----------------------------------------------------------
void OsgDolfinObject::updateBoundaryIndicatorColors()
//----------------------------------------------------------
{
  for(dolfin::FaceIterator v(dolfin_object.surface_mesh()); !v.end(); ++v)
  {
    int indicator = dolfin_object.getFaceBoundaryIndicator(v->index());
    
    if(indicator > 8)
      indicator = 8;
    
    (*face_colors_)[v->index()] = (*color_palette_)[indicator];
  }
  mesh_->setColorArray(face_colors_.get());
  
  for(dolfin::EdgeIterator v(dolfin_object.surface_mesh()); !v.end(); ++v)
  {
    int indicator = dolfin_object.getEdgeBoundaryIndicator(v->index());
    
    if(indicator > 8)
      indicator = 8;
    
    (*edge_colors_)[v->index()] = (*color_palette_)[indicator];
  }
  edge_->setColorArray(edge_colors_.get());
  
  for(dolfin::VertexIterator v(dolfin_object.surface_mesh()); !v.end(); ++v)
  {
    int indicator = dolfin_object.getVertexBoundaryIndicator(v->index());
    
    if(indicator > 8)
      indicator = 8;
    
    (*vertex_colors_)[v->index()] = (*color_palette_)[indicator];
  }
  vertex_->setColorArray(vertex_colors_.get());
}
//----------------------------------------------------------
uint OsgDolfinObject::findEdgeFromVertices(uint id1, uint id2)
//----------------------------------------------------------
{
  // Find the edge that is built up of the two vertices id1 and id2.

  // Brute force: iterate through the edges until found.
  // Maybe there is a more elegant solution out there

  for (dolfin::EdgeIterator f(dolfin_object.surface_mesh()); !f.end(); ++f)
  {
    const uint* entities = f->entities(0);
    uint  num_entities = f->num_entities(0);
    uint  index = f->index();

    if(num_entities == 2)
    {
      if( (entities[0] == id1 && entities[1] == id2) ||
          (entities[1] == id1 && entities[0] == id2))
        return index;
    }
  }
  return 0;
}


//----------------------------------------------------------
uint OsgDolfinObject::findFaceFromVertices(uint id1, uint id2, uint id3)
//----------------------------------------------------------
{
  // Find the face that is built up of the three vertices id1, id2 and id3.

  // Brute force: iterate through the faces until found.
  // Maybe there is a more elegant solution out there

  for (dolfin::FaceIterator f(dolfin_object.surface_mesh()); !f.end(); ++f)
  {
    const uint* entities = f->entities(0);
    uint  num_entities = f->num_entities(0);
    uint  index = f->index();
    
    if(num_entities == 3)
    {
      if( entities[0] == id1 && entities[1] == id2 && entities[2] == id3 )
	return index;
    }
  }
  return 0;
}
