#include <scenegraph/OsgDrawableObject.h>

#include <osg/BlendFunc>
#include <osg/BoundingSphere>
#include <osg/CullFace>
#include <osg/Geode>
#include <osg/Group>
#include <osg/LightModel>
#include <osg/LineStipple>
#include <osg/Material>
#include <osg/Notify>
#include <osg/PolygonMode>
#include <osg/PolygonOffset>
#include <osg/PrimitiveSet>
#include <osg/ShadeModel>
#include <osg/Shape>
#include <osg/ShapeDrawable>
#include <osgFX/Scribe>
#include <osgUtil/Optimizer>

#include <QObject>
#include <QDebug>
#include <iostream>

#include <cmath>


//----------------------------------------------------------
OsgDrawableObject::OsgDrawableObject()
  : OsgTransformObject(),
    data_is_ok_(false),
    old_render_mode_(-1)
//----------------------------------------------------------
{
  default_simres_state_ = new SimResState();
  current_simres_state_ = default_simres_state_;
}

//----------------------------------------------------------
OsgDrawableObject::OsgDrawableObject(const OsgDrawableObject& obj)
  : OsgTransformObject(obj),
    data_is_ok_(obj.data_is_ok_),
    old_render_mode_(obj.old_render_mode_)
//----------------------------------------------------------
{
  if(obj.default_simres_state_.valid())
    default_simres_state_ = new SimResState(*(obj.default_simres_state_.get()));

  current_simres_state_ = default_simres_state_;
}
	   
//----------------------------------------------------------
OsgDrawableObject::~OsgDrawableObject()
//----------------------------------------------------------
{
  // Do nothing
}
//----------------------------------------------------------
void OsgDrawableObject::updateBoundingSphere()
//----------------------------------------------------------
{
  osg::Sphere *sphere = 0;
  if (!bounding_sphere_.valid()) {
    geode_sphere_ = new osg::Geode();
    addChild(geode_sphere_.get());
    
    sphere = new osg::Sphere();
    bounding_sphere_ = new osg::ShapeDrawable(sphere);
    bounding_sphere_->setColor(osg::Vec4(0.8, 0.8, 0.8, 0.5));
    
    osg::StateSet* stateset = bounding_sphere_->getOrCreateStateSet();
    stateset->setMode(GL_BLEND,osg::StateAttribute::ON);
    
    //osg::BlendFunc* bf = new osg::BlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    //stateset->setAttribute(bf);
    
    stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
    
    geode_sphere_->addDrawable(bounding_sphere_.get());
  } else  {
    sphere = dynamic_cast<osg::Sphere*>(bounding_sphere_->getShape());
    if (!geode_sphere_->containsDrawable(bounding_sphere_.get()))
      geode_sphere_->addDrawable(bounding_sphere_.get());
  }
  osg::BoundingSphere bound = computeBound();
  sphere->setCenter(bound.center());
  sphere->setRadius(bound.radius());
}

//----------------------------------------------------------
void OsgDrawableObject::updateBoundingBox()
//----------------------------------------------------------
{
  std::cerr << "OsgDrawable::updateBoundingBox() " << std::endl;

  if (!bbox_trans_.valid()) 
  {
    bbox_trans_ = new osg::MatrixTransform();
    bbox_trans_->setName("bbox_trans");
    addChild(bbox_trans_.get());

    if (!bbox_geode_.valid()) 
    {
      bbox_geode_ = new osg::Geode();

      bbox_geom_ = new osg::Geometry();
      bbox_geode_->addDrawable(bbox_geom_.get());

      osg::Vec3Array *vertices = new osg::Vec3Array(8);
      (*vertices)[0] = osg::Vec3(-0.5, -0.5, -0.5);
      (*vertices)[1] = osg::Vec3( 0.5, -0.5, -0.5);
      (*vertices)[2] = osg::Vec3( 0.5,  0.5, -0.5);
      (*vertices)[3] = osg::Vec3(-0.5,  0.5, -0.5);
      (*vertices)[4] = osg::Vec3(-0.5, -0.5,  0.5);
      (*vertices)[5] = osg::Vec3( 0.5, -0.5,  0.5);
      (*vertices)[6] = osg::Vec3( 0.5,  0.5,  0.5);
      (*vertices)[7] = osg::Vec3(-0.5,  0.5,  0.5);
	  
      bbox_geom_->setVertexArray(vertices);

      osg::StateSet *stateset = bbox_geom_->getOrCreateStateSet();
	  
      osg::LineWidth *linewidth = new osg::LineWidth(1.0);
      stateset->setAttribute(linewidth, osg::StateAttribute::ON);

      stateset->setMode(GL_LIGHTING,osg::StateAttribute::OFF);

      bbox_colors_ = new osg::Vec4Array();
      bbox_colors_->push_back(osg::Vec4(1.0, 1.0, 0.0, 1.0));
	  
      bbox_geom_->setColorArray(bbox_colors_.get());
      bbox_geom_->setColorBinding(osg::Geometry::BIND_OVERALL);

      osg::DrawElementsUInt *indices = new osg::DrawElementsUInt(osg::PrimitiveSet::LINES, 0);
      indices->push_back(0); indices->push_back(1);
      indices->push_back(1); indices->push_back(2);
      indices->push_back(2); indices->push_back(3);
      indices->push_back(3); indices->push_back(0);
      indices->push_back(4); indices->push_back(5);
      indices->push_back(5); indices->push_back(6);
      indices->push_back(6); indices->push_back(7);
      indices->push_back(7); indices->push_back(4);

      indices->push_back(0); indices->push_back(4);
      indices->push_back(1); indices->push_back(5);
      indices->push_back(2); indices->push_back(6);
      indices->push_back(3); indices->push_back(7);
	  
      bbox_geom_->addPrimitiveSet(indices);
    }
    bbox_trans_->addChild(bbox_geode_.get());
  } else {
    if (!containsNode(bbox_trans_.get()))
      addChild(bbox_trans_.get());
  }

  // Update bouding box geometry.
  osg::BoundingBox bbox = getBoundingBox();
      
  osg::Matrix Mt, Ms;
  Mt.makeTranslate(bbox.center());
  Ms.makeScale(osg::Vec3(bbox.xMax()-bbox.xMin(), 
			 bbox.yMax()-bbox.yMin(), 
			 bbox.zMax()-bbox.zMin()));

  bbox_trans_->setMatrix( Ms * Mt );

  // Update bounding box color
  const SimResState &current_state = getCurrentState();
  double r = current_state.getHighlightColor().red()  /255.0;
  double g = current_state.getHighlightColor().green()/255.0;
  double b = current_state.getHighlightColor().blue() /255.0;
  (*bbox_colors_)[0] = osg::Vec4(r, g, b, 1.0);
  bbox_geom_->dirtyDisplayList();
}
//----------------------------------------------------------
void OsgDrawableObject::updateAxes() 
//----------------------------------------------------------
{
  std::cerr << "OsgDrawable::updateAxes() " << std::endl;

  if (!axes.valid()) 
  {

    // Create and add the root node of the axes view
    axes = new osg::Switch;
    axes->setName("axes_switch");
    addChild(axes);

    // Create the transform node which will scale the axes
    axes_transform = new osg::MatrixTransform();
    axes_transform->setName("axes_trans");
    axes->addChild(axes_transform.get(), true);

    // Create the leaf which hold the geometry
    osg::ref_ptr<osg::Geode> axes_geode = new osg::Geode();
    axes_transform->addChild(axes_geode);

    // Create the geometry and add it to the leaf node
    axes_geom = new osg::Geometry();
    axes_geode->addDrawable(axes_geom.get());

    osg::Vec3Array* vertices = new osg::Vec3Array(6);
    // X
    (*vertices)[0] = osg::Vec3(-1.0, .0, .0);
    (*vertices)[1] = osg::Vec3( 1.0, .0, .0);

    // Y
    (*vertices)[2] = osg::Vec3(.0, -1.0, .0);
    (*vertices)[3] = osg::Vec3(.0,  1.0,  .0);

    // Z
    (*vertices)[4] = osg::Vec3( .0, .0, -1.0);
    (*vertices)[5] = osg::Vec3( .0, .0,  1.0);

    
    axes_geom->setVertexArray(vertices);

    // Set the line width
    osg::StateSet *stateset = axes_geom->getOrCreateStateSet();
    osg::LineWidth *linewidth = new osg::LineWidth(3.5);
    stateset->setAttribute(linewidth, osg::StateAttribute::ON);

    // Turn off lightning for the axes
    stateset->setMode(GL_LIGHTING, osg::StateAttribute::OFF);

    // Create the colors
    axes_colors = new osg::Vec4Array();
    //Todo: This should definitely not be the default color...
    axes_colors->push_back(osg::Vec4(1.0, 1.0, 0.0, 1.0));
	  
    axes_geom->setColorArray(axes_colors.get());
    axes_geom->setColorBinding(osg::Geometry::BIND_OVERALL);

    osg::DrawElementsUInt* indices = new osg::DrawElementsUInt(osg::PrimitiveSet::LINES, 0);
    indices->push_back(0); indices->push_back(1);
    indices->push_back(2); indices->push_back(3);
    indices->push_back(4); indices->push_back(5);
	  
    axes_geom->addPrimitiveSet(indices);
    
    // Add the cones indicating axis direction
    osg::ref_ptr<osg::ShapeDrawable> cone = 
      new osg::ShapeDrawable(new osg::Cone(osg::Vec3(.0, .0, .0), .005, .01));
    cone->setColor(osg::Vec4(1.0, 1.0, 0.0, 1.0));

    {
      // X cone
      osg::ref_ptr<osg::MatrixTransform> x_transform = new osg::MatrixTransform;
      osg::Matrix Mt, Mr;
      Mt.makeTranslate(osg::Vec3(1.0, .0, .0));
      Mr.makeRotate(M_PI/2, osg::Vec3(.0, 1.0, .0));
      x_transform->setMatrix(Mr * Mt);

      osg::ref_ptr<osg::Geode> x_cone_geode = new osg::Geode;

      axes_transform->addChild(x_transform);
      x_transform->addChild(x_cone_geode);
      x_cone_geode->addDrawable(cone);
    }

    {
      // Y cone
      osg::ref_ptr<osg::MatrixTransform> y_transform = new osg::MatrixTransform;
      osg::Matrix Mt, Mr;
      Mt.makeTranslate(osg::Vec3(.0, 1.0, .0));
      Mr.makeRotate(3*M_PI/2, osg::Vec3(1.0, .0, .0));
      y_transform->setMatrix(Mr * Mt);

      osg::ref_ptr<osg::Geode> y_cone_geode = new osg::Geode;

      axes_transform->addChild(y_transform);
      y_transform->addChild(y_cone_geode);
      y_cone_geode->addDrawable(cone);
    }
    

    {
      // Z cone
      osg::ref_ptr<osg::MatrixTransform> z_transform = new osg::MatrixTransform;
      osg::Matrix Mt;
      Mt.makeTranslate(osg::Vec3(.0, 0.0, 1.0));
      z_transform->setMatrix(Mt);
      
      osg::ref_ptr<osg::Geode> z_cone_geode = new osg::Geode;
      
      axes_transform->addChild(z_transform);
      z_transform->addChild(z_cone_geode);
      z_cone_geode->addDrawable(cone);
    }
  }

  // Scale the axes according to the bounding box
  osg::BoundingBox bbox = getBoundingBox();


  // TODO: More work needed here
  // If a mesh is not centered around origo, the axes will
  // "point out".
  // Hoewever the mesh might not be centered around origo
  osg::Matrix Ms;
  Ms.makeScale(osg::Vec3(std::max(bbox.xMax(), std::abs(bbox.xMin())) * 1.1, 
                         std::max(bbox.yMax(), std::abs(bbox.yMin())) * 1.1, 
                         std::max(bbox.zMax(), std::abs(bbox.zMin())) * 1.1));

  axes_transform->setMatrix( Ms );
  /*
  // Update axes color
  const SimResState &current_state = getCurrentState();
  double r = current_state.getHighlightColor().red()  /255.0;
  double g = current_state.getHighlightColor().green()/255.0;
  double b = current_state.getHighlightColor().blue() /255.0;
  (*axes_colors)[0] = osg::Vec4(r, g, b, 1.0);
  */


  if (current_simres_state_->getShowAxes()) 
  {
    axes->setAllChildrenOn();
  }
  else 
  {
    axes->setAllChildrenOff();
  }


  axes_geom->dirtyDisplayList();
  
}
//----------------------------------------------------------
void OsgDrawableObject::updateGroundPlane()
//----------------------------------------------------------
{
  std::cerr << "OsgDrawable::updateGroundPlane" << std::endl;

  if (!ground_plane.valid()) 
  {
    ground_plane = new osg::Switch;
    ground_plane->setName("ground_plane");
    addChild(ground_plane.get());


    // Create the transformation used to scale the geometry
    ground_transform = new osg::MatrixTransform;
    ground_plane->addChild(ground_transform.get(), true);

    // Create the geometry
    ground_geom = new osg::Geometry;

    osg::Vec3Array* vertices       = new osg::Vec3Array(20);
    osg::DrawElementsUInt* indices = new osg::DrawElementsUInt(osg::PrimitiveSet::LINES, 0);

    for (uint i = 0; i < 10; i++) 
    {
      //Add the two vertices
      (*vertices)[i*2]      = osg::Vec3(-1.0,  (i+1)/10.0, .0); 
      (*vertices)[i*2 + 1]  = osg::Vec3( 1.0,  (i+1)/10.0, .0);

      //Then define the line
      indices->push_back(i*2); 
      indices->push_back(i*2 + 1);
    }

    ground_geom->setVertexArray(vertices);
    ground_geom->addPrimitiveSet(indices);

    ground_colors = new osg::Vec4Array;
    ground_colors->push_back(osg::Vec4(1.0, 1.0, 0.0, 1.0));	  
    ground_geom->setColorArray(ground_colors.get());
    ground_geom->setColorBinding(osg::Geometry::BIND_OVERALL);

    osg::StateSet *stateset = ground_geom->getOrCreateStateSet();
    osg::LineWidth *linewidth = new osg::LineWidth(1.5);
    stateset->setAttribute(linewidth, osg::StateAttribute::ON);

    // Turn off lightning for the axes
    stateset->setMode(GL_LIGHTING, osg::StateAttribute::OFF);

    //create the geode
    osg::ref_ptr<osg::Geode> geode = new osg::Geode;
    geode->addDrawable(ground_geom);

    
    {
      ground_transform->addChild(geode.get());
    }

    {
      osg::ref_ptr<osg::MatrixTransform> x_minus = new osg::MatrixTransform;
      ground_transform->addChild(x_minus.get());
      osg::Matrix Mr;
      Mr.makeRotate( M_PI, osg::Vec3(.0, .0, 1.0));
      x_minus->setMatrix(Mr);
      x_minus->addChild(geode.get());
    }    

    
    {
      //create the geode and the rotation holding the indicators in the y direction
      osg::ref_ptr<osg::MatrixTransform> y_rotation = new osg::MatrixTransform;
      osg::Matrix Mr;
      Mr.makeRotate( M_PI/2, osg::Vec3(.0, .0, 1.0));
      y_rotation->setMatrix(Mr);

      y_rotation->addChild(geode);
      ground_transform->addChild(y_rotation);
    }

    {
      // And the minus part of the y direction indicators
      osg::ref_ptr<osg::MatrixTransform> y_minus = new osg::MatrixTransform;
      osg::Matrix Mr;
      Mr.makeRotate( 3*M_PI/2, osg::Vec3(.0, .0, 1.0));
      y_minus->setMatrix(Mr);
      y_minus->addChild(geode);
      ground_transform->addChild(y_minus);
    }
    
    
  }


  // Scale the plane indicators according to the bounding box
  osg::BoundingBox bbox = getBoundingBox();

  osg::Matrix Ms;
  Ms.makeScale(osg::Vec3(std::max(bbox.xMax(), std::abs(bbox.xMin())), 
			 std::max(bbox.yMax(), std::abs(bbox.yMin())), 
			 0));

  ground_transform->setMatrix(Ms);


  if (current_simres_state_->getShowPlane()) 
  {
    ground_plane->setAllChildrenOn();
  }
  else 
  {
    ground_plane->setAllChildrenOff();
  }

  ground_geom->dirtyDisplayList();
  

}
//----------------------------------------------------------
void OsgDrawableObject::setCurrentState(SimResState* state)
//----------------------------------------------------------
{
  std::cout << "OsgDrawable::setCurrentState" << std::endl;

  current_simres_state_ = state;

  // Update
  updateRenderMode();
  updateTransparency();
  updateColorAndIntensity();
  updateSortMode();
  updateCullMode();
  updateTwoSidedLighting();
  updateBaseColor();
  updateWireframeColor();
}

//----------------------------------------------------------
void OsgDrawableObject::applyState(const SimResState& s)
//----------------------------------------------------------
{
  // std::cerr << "OsgDrawableObject::applyState: " << this
  //	 << " " << &s << std::endl;

  // Set values, keep track of which update
  // methods will be required later.

  bool update_transparency         = false;
  bool update_colors_and_intensity = false;
  bool update_render_mode          = false;
  bool update_sort_mode            = false;
  bool update_cull_mode            = false;
  bool update_two_sided            = false;
  bool update_base_color           = false;
  bool update_wireframe_color      = false;

  SimResState& c = *(current_simres_state_);

  // Avoid sending signals during apply.
  current_simres_state_->blockSignals(true);

  if(c.setTransparency(s.getTransparency()))
    update_transparency=true;

  if(c.setDiffuseIntensity(s.getDiffuseIntensity()))
    update_colors_and_intensity=true;

  if(c.setAmbientIntensity(s.getAmbientIntensity()))
    update_colors_and_intensity=true;

  if(c.setEmissionIntensity(s.getEmissionIntensity()))
    update_colors_and_intensity=true;

  if(c.setSpecularIntensity(s.getSpecularIntensity()))
    update_colors_and_intensity=true;

  if(c.setShininess(s.getShininess()))
    update_colors_and_intensity=true;

  if(c.setRenderMode(s.getRenderMode()))
    update_render_mode=true;

  if(c.setSortMode(s.getSortMode()))
    update_sort_mode=true;

  if(c.setCullMode(s.getCullMode()))
    update_cull_mode=true;

  if(c.setTwoSidedLighting(s.getTwoSidedLighting()))
    update_two_sided=true;

  if(c.setBaseColor(s.getBaseColor()))
    update_base_color=true;

  if(c.setBackColor(s.getBackColor()))
    update_base_color=true;

  if(c.setWireframeColor(s.getWireframeColor()))
    update_wireframe_color=true;

  // Start sending signals again.
  current_simres_state_->blockSignals(false);

  // Call update methods.

  if(update_transparency)         updateTransparency();
  if(update_colors_and_intensity) updateColorAndIntensity();
  if(update_render_mode)          updateRenderMode();
  if(update_sort_mode)            updateSortMode();
  if(update_cull_mode)            updateCullMode();
  if(update_two_sided)            updateTwoSidedLighting();
  if(update_base_color)           updateBaseColor();
  if(update_wireframe_color)      updateWireframeColor();
}
    
//----------------------------------------------------------
SimResState& OsgDrawableObject::getDefaultState()
//----------------------------------------------------------
{
  return *(default_simres_state_);
}
    
//----------------------------------------------------------
const SimResState& OsgDrawableObject::getDefaultState() const
//----------------------------------------------------------
{
  return *(default_simres_state_);
}
    
//----------------------------------------------------------
SimResState& OsgDrawableObject::getCurrentState() 
//----------------------------------------------------------
{
  return *(current_simres_state_);
}
    
//----------------------------------------------------------
const SimResState& OsgDrawableObject::getCurrentState() const
//----------------------------------------------------------
{
  return *(current_simres_state_);
}

//----------------------------------------------------------
osg::Geode* OsgDrawableObject::getGeode() 
//----------------------------------------------------------
{
  return geode_.get();
}
    
//----------------------------------------------------------
const osg::Geode* OsgDrawableObject::getGeode() const
//----------------------------------------------------------
{
  return geode_.get();
}

//----------------------------------------------------------
void OsgDrawableObject::initialize()
//----------------------------------------------------------
{
  std::cerr << "OsgDrawableObject::initialize" << std::endl;

  OsgTransformObject::initialize();

  // Make sure that we have a state object.
  
  //if(!default_simres_state_.valid())
  //  default_simres_state_ = new SimResState();
   
  //current_simres_state_ = default_simres_state_;
   
  //Set the current state
  //current_simres_state_ = simres_state_;


  // Connect to signals from the state object.
  QObject::connect(current_simres_state_.get(), SIGNAL(stateChanged(int)),
                   this, SLOT(stateChanged(int)));
   
  // Allocate materials, states, etc.
   
  material_  = new osg::Material;
      
  cull_face_ = new osg::CullFace();
      
  double transparency = current_simres_state_->getTransparency();
   
  colors_ = new osg::Vec4Array();
  colors_->push_back(osg::Vec4(1.0, 0.0, 1.0, 1.0-transparency));

  // Use it only if necessary (in derived class)
  //secondary_colors_ = new osg::Vec4Array();
  //secondary_colors_->push_back(osg::Vec4(1.0, 0.0, 0.0, 1.0-transparency));
}

//----------------------------------------------------------
void OsgDrawableObject::finalize()
//----------------------------------------------------------
{
  OsgTransformObject::finalize();

  if(geode_==NULL) return;

  // run optimization over the scene graph
  //    osgUtil::Optimizer optimzer;
  //    optimzer.optimize(geode_.get());	

  if(scribe_.get()==NULL)
  {
    scribe_ = new osgFX::Scribe();
    scribe_->addChild(geode_.get());
  }
      
  // Turn blending on
  osg::StateSet* stateset= geode_->getOrCreateStateSet();
  stateset->setMode(GL_BLEND,osg::StateAttribute::ON);

  osg::BlendFunc* bf = new osg::BlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  stateset->setAttribute(bf);
}

//----------------------------------------------------------
void OsgDrawableObject::updateTransparency()
//----------------------------------------------------------
{
  osg::Vec4Array* vec = colors_.get();

  //    std::cerr << "OsgDrawableObject::updateTransparency" << std::endl;

  if(vec) 
  {
    double transparency = current_simres_state_->getTransparency();
	  
    if(geode_.get()) 
    {
      osg::StateSet* stateset = geode_->getOrCreateStateSet();
	  
      // add rendering hint
      if(transparency != 0.0)
	stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
      else
	stateset->setRenderingHint(osg::StateSet::OPAQUE_BIN);
    }
	
    osg::Vec4Array::iterator it;
    for(it=vec->begin() ; it != vec->end() ; it++)
      it->w() = 1.0-transparency;
	
    ambient_.w()  = 1.0-transparency;
    diffuse_.w()  = 1.0-transparency;
    specular_.w() = 1.0-transparency;
    emission_.w() = 1.0-transparency;
	
    updateMaterial();
  }
}

//----------------------------------------------------------
void OsgDrawableObject::updateColorAndIntensity()
//----------------------------------------------------------
{
  if(!data_is_ok_) return;
      
  //    std::cerr << "OsgDrawableObject::updateColorAndIntensity" << std::endl;
      
  double diffuse_intensity  = current_simres_state_->getDiffuseIntensity();
  double ambient_intensity  = current_simres_state_->getAmbientIntensity();
  double emission_intensity = current_simres_state_->getEmissionIntensity();
  double specular_intensity = current_simres_state_->getSpecularIntensity();
  double transparency       = current_simres_state_->getTransparency();

  diffuse_  = osg::Vec4(diffuse_intensity,  diffuse_intensity,  diffuse_intensity,  1.0-transparency);
  ambient_  = osg::Vec4(ambient_intensity,  ambient_intensity,  ambient_intensity,  1.0-transparency);
  emission_ = osg::Vec4(emission_intensity, emission_intensity, emission_intensity, 1.0-transparency);
  specular_ = osg::Vec4(specular_intensity, specular_intensity, specular_intensity, 1.0-transparency);

  updateMaterial();
}

//----------------------------------------------------------
void OsgDrawableObject::updateMaterial()
//----------------------------------------------------------
{
  //    std::cerr << "OsgDrawableObject::updateMaterial" << std::endl;

  const SimResState &current_state = getCurrentState();

  double shininess = current_state.getShininess();

  osg::Vec4 scolor;

  scolor.x() = current_state.getBackColor().red()  /255.0;
  scolor.y() = current_state.getBackColor().green()/255.0;
  scolor.z() = current_state.getBackColor().blue() /255.0;
  scolor.w() = 1.0;

  material_->setAmbient  (osg::Material::FRONT_AND_BACK, ambient_);
  material_->setDiffuse  (osg::Material::FRONT_AND_BACK, diffuse_);
  material_->setSpecular (osg::Material::FRONT_AND_BACK, specular_);
  material_->setEmission (osg::Material::FRONT_AND_BACK, emission_);
  material_->setShininess(osg::Material::FRONT_AND_BACK, shininess);

  material_->setAmbient  (osg::Material::BACK, ambient_);
  material_->setDiffuse  (osg::Material::BACK, diffuse_);
  material_->setSpecular (osg::Material::BACK, specular_);
  material_->setEmission (osg::Material::BACK, scolor); //emission_);
  material_->setShininess(osg::Material::BACK, shininess);

  material_->setColorMode(osg::Material::AMBIENT);
  //material_->setColorMode(osg::Material::OFF);

  if (geode_.valid()) {
    osg::StateSet* stateset = geode_->getOrCreateStateSet();
    stateset->setAttributeAndModes(material_.get(), osg::StateAttribute::ON);
    stateset->setMode(GL_LIGHTING,osg::StateAttribute::ON);
  }
}

//----------------------------------------------------------
void OsgDrawableObject::updateCullMode()
//----------------------------------------------------------
{
  if(!data_is_ok_) return;

  //    std::cerr << "OsgDrawableObject::updateCullMode" << std::endl;

  int culling = current_simres_state_->getCullMode();

  // Initialize to culling off (culling==0), with mode set to BACK (culling==1)
  osg::StateAttribute::GLModeValue val = osg::StateAttribute::OFF;
  osg::CullFace::Mode mode = osg::CullFace::BACK;

  if(culling > 0)   val = osg::StateAttribute::ON;
  if(culling == 2) mode = osg::CullFace::FRONT;

  cull_face_->setMode(mode);

  osg::StateSet* stateset = geode_->getOrCreateStateSet();
  stateset->setAttributeAndModes(cull_face_.get(), val);
  stateset->setMode(GL_CULL_FACE, val);
}

//----------------------------------------------------------
void OsgDrawableObject::updateTwoSidedLighting()
//----------------------------------------------------------
{
  if(!data_is_ok_) return;

  osg::StateSet* stateset = geode_->getOrCreateStateSet();

  // turn two sided lighting on
  osg::LightModel* lm = new osg::LightModel();
  lm->setTwoSided(current_simres_state_->getTwoSidedLighting());
  stateset->setAttribute(lm);
}

//----------------------------------------------------------
void OsgDrawableObject::updateRenderMode()
//----------------------------------------------------------
{
  //    if(!data_is_ok_) return;

  //qDebug() << "OsgDrawableObject::updateRenderMode";
   
  if(!current_simres_state_ || !geode_.valid())
    return;

  int render_mode = current_simres_state_->getRenderMode();

  // Check if old mode uses Scribe node.
  // In that case we need to modify the scene graph.
    
  if(old_render_mode_==4 || old_render_mode_==5) {
    replaceChild(scribe_.get(), geode_.get());
  }

  osg::StateSet* stateset = geode_->getOrCreateStateSet();

  osg::PolygonMode* polymode = dynamic_cast<osg::PolygonMode*>
    (stateset->getAttribute(osg::StateAttribute::POLYGONMODE));
  if (!polymode) polymode = new osg::PolygonMode;

  osg::ShadeModel* sm = dynamic_cast<osg::ShadeModel*>
    (stateset->getAttribute(osg::StateAttribute::SHADEMODEL));
  if (!sm) sm = new osg::ShadeModel();

  // Change the state attribute settings.
  if(render_mode == SimResState::SMOOTH) 
  {
    polymode->setMode(osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::FILL);
    sm->setMode(osg::ShadeModel::SMOOTH);
  }
  else if(render_mode == SimResState::FLAT) 
  {
    polymode->setMode(osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::FILL);
    sm->setMode(osg::ShadeModel::FLAT);
  }
  else if(render_mode == SimResState::POINTS) 
  {
    polymode->setMode(osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::POINT);
  }
  else if(render_mode == SimResState::WIREFRAME) 
  {
    polymode->setMode(osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::LINE);
  }
  else if(render_mode == SimResState::WIRE_AND_FLAT) 
  {
    // Wireframe only works when optimizer is off.
    polymode->setMode(osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::FILL);
    sm->setMode(osg::ShadeModel::FLAT);
    replaceChild(geode_.get(), scribe_.get());
  }
  else if(render_mode == SimResState::WIRE_AND_SMOOTH) 
  {
    // Wireframe only works when optimizer is off.
    polymode->setMode(osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::FILL);
    sm->setMode(osg::ShadeModel::SMOOTH);
    replaceChild(geode_.get(), scribe_.get());
  }
      
  stateset->setAttributeAndModes(polymode,osg::StateAttribute::ON);
  stateset->setAttributeAndModes(sm,      osg::StateAttribute::ON);
      
  old_render_mode_ = render_mode;
}

//----------------------------------------------------------
void OsgDrawableObject::updateSortMode()
//----------------------------------------------------------
{
  if(!data_is_ok_) return;
      
  // std::cerr << "OsgDrawableObject::updateSortMode" << std::endl;
    
  //int sort_mode = current_simres_state_->getSortMode();
}

//----------------------------------------------------------
void OsgDrawableObject::addChildObject(SimResState* child, const QString&)
//----------------------------------------------------------
{
  //SimResState* state = dynamic_cast<SimResState*>(child);

  //if(state) {
  if(default_simres_state_.valid())
    std::cerr << "Warning: trying to replace default state object." << std::endl;
	
  default_simres_state_ = child;
  setCurrentState(child);
    //}
}

//----------------------------------------------------------
void OsgDrawableObject::stateChanged(int id)
//----------------------------------------------------------
{
  switch(id) {
  case SimResState::RenderMode:
    updateRenderMode();
    break;
	
  case SimResState::SortMode:
    updateSortMode();
    break;
	
  case SimResState::CullMode:
    updateCullMode();
    break;
	
  case SimResState::TwoSidedLighting:
    updateTwoSidedLighting();
    break;
	
  case SimResState::Transparency:
    updateTransparency();
    break;
	
  case SimResState::DiffuseIntensity:
  case SimResState::AmbientIntensity:
  case SimResState::EmissionIntensity:
  case SimResState::SpecularIntensity:
  case SimResState::Shininess:
    updateColorAndIntensity();
    break;
	
  case SimResState::BaseColor:
  case SimResState::BackColor:
    updateBaseColor();
    break;

  case SimResState::Highlight:
  case SimResState::HighlightColor:
    updateHighlight();
    updateBaseColor();
    break;

  case SimResState::WireframeColor:
    updateWireframeColor();
    break;
  case SimResState::Axes :
    updateAxes();
    break;
  case SimResState::Plane :
    updateGroundPlane();
    break;
  }
}

//----------------------------------------------------------
osg::BoundingBox OsgDrawableObject::getBoundingBox() const
//----------------------------------------------------------
{
  if(geode_.get()){
    return geode_->getBoundingBox();
  }

  // Return default bounding box
  return osg::BoundingBox();
}

//----------------------------------------------------------
void OsgDrawableObject::updateBaseColor()
//----------------------------------------------------------
{
  //printf("OsgDrawableObject::updateBaseColor\n");

  if(colors_.get()) {
    osg::Vec4 color, scolor;
    if (0) { //current_simres_state_->getHighlight()) {
      color.x() = current_simres_state_->getHighlightColor().red()/255.0;
      color.y() = current_simres_state_->getHighlightColor().green()/255.0;
      color.z() = current_simres_state_->getHighlightColor().blue()/255.0;
      color.w() = 1.0 - current_simres_state_->getTransparency();
      scolor = color;
    }
    else {
      color.x() = current_simres_state_->getBaseColor().red()/255.0;
      color.y() = current_simres_state_->getBaseColor().green()/255.0;
      color.z() = current_simres_state_->getBaseColor().blue()/255.0;
      color.w() = 1.0 - current_simres_state_->getTransparency();

      scolor.x() = current_simres_state_->getBackColor().red()/255.0;
      scolor.y() = current_simres_state_->getBackColor().green()/255.0;
      scolor.z() = current_simres_state_->getBackColor().blue()/255.0;
      scolor.w() = 1.0 - current_simres_state_->getTransparency();
    }
	
    osg::Vec4Array::iterator it;
    for(it=colors_->begin(); it!=colors_->end(); it++) {
      *it = color;
    }
	
    if (secondary_colors_.valid()) {
      for(it=secondary_colors_->begin(); it!=secondary_colors_->end(); it++) {
	*it = scolor;
      }
    }
  }
}

//----------------------------------------------------------
void OsgDrawableObject::updateWireframeColor()
//----------------------------------------------------------
{
  if(scribe_.get()) 
  {
    double r = current_simres_state_->getWireframeColor().red()/255.0;
    double g = current_simres_state_->getWireframeColor().green()/255.0;
    double b = current_simres_state_->getWireframeColor().blue()/255.0;
	
    scribe_->setWireframeColor(osg::Vec4(r, g, b, 1.0));
  }
}

//----------------------------------------------------------
void OsgDrawableObject::updateTransformObject()
//----------------------------------------------------------
{
  OsgTransformObject::updateTransformObject();

  if (/*!trans_obj_.valid() ||*/ !geode_.valid())
    return;

  //if (!simres_state_.valid()) return;
      
  if (!current_simres_state_->isBoundingSphereShown()) 
  {
    if (bounding_sphere_.valid())
      geode_sphere_->removeDrawable(bounding_sphere_.get());
  }
  else 
    updateBoundingSphere();

}

//----------------------------------------------------------
void OsgDrawableObject::updateHighlight()
//----------------------------------------------------------
{
  if (!current_simres_state_->getHighlight()) {
    if (bbox_trans_.valid())
      removeChild(bbox_trans_.get());
  }
  else 
    updateBoundingBox();
}

//----------------------------------------------------------
osg::BoundingSphere OsgDrawableObject::computeBound() const
//----------------------------------------------------------
{
  if (geode_.valid())
    return geode_->computeBound();
  else
    return OsgTransformObject::computeBound();
}

//----------------------------------------------------------
const bool &OsgDrawableObject::isDataOk() const
//----------------------------------------------------------
{
  return data_is_ok_;
}

//----------------------------------------------------------
void OsgDrawableObject::setDataOk(const bool onoff)
//----------------------------------------------------------
{
  data_is_ok_ = onoff;
}

//----------------------------------------------------------
osg::Vec4 &OsgDrawableObject::diffuse()
//----------------------------------------------------------
{
  return diffuse_;
}

//----------------------------------------------------------
osg::Vec4 &OsgDrawableObject::ambient() 
//----------------------------------------------------------
{
  return ambient_;
}

//----------------------------------------------------------
osg::Vec4 &OsgDrawableObject::emission() 
//----------------------------------------------------------
{
  return emission_;
}

//----------------------------------------------------------
osg::Vec4 &OsgDrawableObject::specular() 
//----------------------------------------------------------
{
  return specular_;
}
