#include <scenegraph/OsgTransformObject.h>

#include <QStringList>
#include <iostream>

//----------------------------------------------------------
OsgTransformObject::OsgTransformObject()
  :euler_x_angle_(0),
   euler_y_angle_(0),
   euler_z_angle_(0),
   euler_rotation_mode_(false),
   object_manipulation_mode_(false)
//----------------------------------------------------------
{
  setName("OsgTransformObject");

  //From TransformObject
  translation_.set(0,0,0);
  scale_.set(1,1,1);
  rot_axis_.set(0,0,1);
  
  rot_angle_ = 0;
  std::cout << "Constructing OsgTransformObject " << std::endl;
  so_.zeroRotation();
}

//----------------------------------------------------------
OsgTransformObject::OsgTransformObject(const OsgTransformObject& obj)
  : osg::MatrixTransform(), //CHECK: Is this correct?
    euler_x_angle_(obj.euler_x_angle_),
    euler_y_angle_(obj.euler_y_angle_),
    euler_z_angle_(obj.euler_z_angle_),
    euler_rotation_mode_(obj.euler_rotation_mode_),
    object_manipulation_mode_(obj.object_manipulation_mode_)
//----------------------------------------------------------
{
}

//----------------------------------------------------------
OsgTransformObject::~OsgTransformObject()
//----------------------------------------------------------
{
}

//----------------------------------------------------------
void OsgTransformObject::setName(QString s)
//----------------------------------------------------------
{
  osg::Node::setName(s.toAscii().constData());
}

///----------------------------------------------------------
const osg::Vec3 &OsgTransformObject::getTranslation() const
///----------------------------------------------------------
{
  return translation_;
}

///----------------------------------------------------------
const osg::Vec3 &OsgTransformObject::getRotationAxis() const
///----------------------------------------------------------
{
  return rot_axis_;
}

///----------------------------------------------------------
const double &OsgTransformObject::getRotationAngle() const
///----------------------------------------------------------
{
  return rot_angle_;
}

///----------------------------------------------------------
const osg::Vec3 &OsgTransformObject::getScale() const
///----------------------------------------------------------
{
  return scale_;
}

///----------------------------------------------------------
void OsgTransformObject::setTranslation(const osg::Vec3 &T) 
//----------------------------------------------------------
{
  translation_ = T;

  if (!object_manipulation_mode_)
    updateMatrix();
}

//----------------------------------------------------------
void OsgTransformObject::setRotationAxis(const osg::Vec3 &R)
//----------------------------------------------------------
{
  rot_axis_ = R;
  
  if (!object_manipulation_mode_)
    updateMatrix();
}

//----------------------------------------------------------
void OsgTransformObject::setRotationAngle(double rot_angle)
//----------------------------------------------------------
{  
  rot_angle_ = rot_angle;

  if (!object_manipulation_mode_) 
  {
    updateMatrix();
  } 
}

//----------------------------------------------------------
void OsgTransformObject::setScale(const osg::Vec3 &S)
//----------------------------------------------------------
{
  scale_ = S;
  
  updateMatrix();
}
//----------------------------------------------------------
void OsgTransformObject::setScaleOrientation(const osg::Quat &so)
//----------------------------------------------------------
{
  so_ = so;
}

//----------------------------------------------------------
const osg::Quat &OsgTransformObject::getScaleOrientation() const
//----------------------------------------------------------
{
  return so_;
}
//----------------------------------------------------------
void OsgTransformObject::setVisible(bool visible)
//----------------------------------------------------------
{
  visible_ = visible;
}

//----------------------------------------------------------
bool OsgTransformObject::getVisible() const
//----------------------------------------------------------
{
  return visible_;
}
//----------------------------------------------------------
void OsgTransformObject::setEulerXAngle(double euler_x_angle)
//----------------------------------------------------------
{
  euler_x_angle_ = euler_x_angle;

  updateMatrix();
}

//----------------------------------------------------------
void OsgTransformObject::setEulerYAngle(double euler_y_angle)
//----------------------------------------------------------
{
  euler_y_angle_ = euler_y_angle;

  updateMatrix();
}

//----------------------------------------------------------
void OsgTransformObject::setEulerZAngle(double euler_z_angle)
//----------------------------------------------------------
{
  euler_z_angle_ = euler_z_angle;

  updateMatrix();
}

//----------------------------------------------------------
void OsgTransformObject::updateEulerAngles(osg::Quat q)
//----------------------------------------------------------
{
  // Convert from quaternion to euler angles

  euler_z_angle_ = atan2(2*q.y()*q.w()-2*q.x()*q.z() , 1 - 2*q.y()*q.y() - 2*q.z()*q.z());
  euler_y_angle_ = asin(2*q.x()*q.y() + 2*q.z()*q.w()); 
  euler_x_angle_ = atan2(2*q.x()*q.w()-2*q.y()*q.z() , 1 - 2*q.x()*q.x() - 2*q.z()*q.z());
}

//----------------------------------------------------------
void OsgTransformObject::updateTransRotScale()
//----------------------------------------------------------
{
  //cerr << "OsgTransformObject::updateTransRotScale..." << endl;
  osg::Matrix M = getMatrix();

  osg::Vec3 translation;
  osg::Vec3 rot_axis;
  double    rot_angle;
  osg::Vec3 scale;
  osg::Quat Qr, so;

  M.decompose(translation, Qr, scale, so);
  Qr.getRotate(rot_angle, rot_axis);

  setTranslation(translation);
  setRotationAxis(rot_axis);
  setRotationAngle(rot_angle);
  setScale(scale);
  setScaleOrientation(so);
}

//----------------------------------------------------------
void OsgTransformObject::updateMatrix()
//----------------------------------------------------------
{
  if (!euler_rotation_mode_) {
    osg::Matrix M, Mt, Ms, Mr, Mso;
    
    Mt.makeTranslate(translation_);
    Ms.makeScale(scale_);
    Mr.makeRotate(osg::Quat(rot_angle_, rot_axis_));
    Mso.makeRotate(so_);
    
    M = Ms * Mso * Mr * Mt;
    
    setMatrix(M);
  }
  else {
    osg::Vec3 local_origo = getBoundingBox().center();
    
    // Translate to local origo (center of BB of the object)
    osg::Matrix M = osg::Matrix::translate(-local_origo);
    
    // Do the scale
    M.postMult(osg::Matrix::scale(scale_));
    
    // ...and rotation
    osg::Matrix rot_mat;
    
    //      printf("euler angles rotation...\n");
    rot_mat = osg::Matrix::rotate(euler_x_angle_, osg::Vec3(1,0,0),
				  euler_y_angle_, osg::Vec3(0,1,0),
				  euler_z_angle_, osg::Vec3(0,0,1));
    
    //      osg::Quat quat = rot_mat.getRotation();
    // Update the rotation axis and angle according to the euler angles
    //      quat.getRotate(rot_angle_, rot_axis_);    
    
    M.postMult(rot_mat);
    
    // Translate back to old position
    M.postMult(osg::Matrix::translate(local_origo));
    
    // Translate to new position
    M.postMult(osg::Matrix::translate(translation_));
    
    setMatrix(M);
    
    // updateTransRotScale();
  }
}

//----------------------------------------------------------
void OsgTransformObject::initialize()
//----------------------------------------------------------
{
}

//----------------------------------------------------------
void OsgTransformObject::finalize()
//----------------------------------------------------------
{
  //   updateMatrix();
}

//----------------------------------------------------------
osg::Matrix OsgTransformObject::getCombinedTransform(osg::Matrix mat_in)
//----------------------------------------------------------
{
  osg::Node::ParentList plist = getParents();
  osg::Node::ParentList::iterator it;

  osg::Matrix mat = osg::Matrix::identity();

  for (it=plist.begin(); it != plist.end(); it++)
    {
      OsgTransformObject* tnode = dynamic_cast<OsgTransformObject*>(*it);
      
      if (tnode)
	mat.preMult(tnode->getMatrix());
    }

  mat.preMult(mat_in);

  return mat;
}

//----------------------------------------------------------
osg::Matrix OsgTransformObject::getCombinedTransform()
//----------------------------------------------------------
{
  return getCombinedTransform(getMatrix());
}

//----------------------------------------------------------
void OsgTransformObject::updateTransformObject()
//----------------------------------------------------------
{
}

//----------------------------------------------------------
osg::Vec3 OsgTransformObject::getTranslation(osg::Matrix mat)
//----------------------------------------------------------
{
  return osg::Vec3(mat(3,0), mat(3,1), mat(3,2));
}

//----------------------------------------------------------
void OsgTransformObject::getAxisAngle(double& angle, osg::Vec3& axis, osg::Matrix mat, osg::Vec3 scale)
//----------------------------------------------------------
{
  //printf("OsgTransformObject::getAxisAngle...\n");


  /*
    printf("getAxisAngle: scale: %lf %lf %lf...\n", scale.x(), scale.y(), scale.z());

    printf("mat before descaling... :\n");
    for (int i=0; i<4; i++)
    {
    for (int j=0; j<4; j++)
    printf("%lf ", mat(i,j));
    printf("\n");
    }
  */

  mat(0,0) /= scale.x();
  mat(0,1) /= scale.x();
  mat(0,2) /= scale.x();

  mat(1,0) /= scale.y();
  mat(1,1) /= scale.y();
  mat(1,2) /= scale.y();

  mat(2,0) /= scale.z();
  mat(2,1) /= scale.z();
  mat(2,2) /= scale.z();

  /*
    printf("mat after descaling... :\n");
    for (int i=0; i<4; i++)
    {
    for (int j=0; j<4; j++)
    printf("%lf ", mat(i,j));
    printf("\n");
    }
  */

  double trace = mat(0,0) + mat(1,1) + mat(2,2) + 1;

  osg::Quat quat;

  if (trace>0)
    {
      double s = 0.5/sqrt(trace);

      quat.w() = 0.25/s;

      quat.x() = (mat(1,2)-mat(2,1))*s;
      quat.y() = (mat(2,0)-mat(0,2))*s;
      quat.z() = (mat(0,1)-mat(1,0))*s;
    }
  else if (mat(0,0) > mat(1,1) && mat(0,0) > mat(2,2))
    {
      double s  = sqrt( 1.0 + mat(0,0) - mat(1,1) - mat(2,2) ) * 2;
      
      quat.x() = 0.5/s;
      quat.y() = (mat(1,0) + mat(0,1) ) / s;
      quat.z() = (mat(2,0) + mat(0,2) ) / s;
      quat.w() = (mat(2,1) + mat(1,2) ) / s;
    }
  else if (mat(1,1) > mat(0,0) && mat(1,1) > mat(2,2) )
    {
      double s  = sqrt( 1.0 + mat(1,1) - mat(0,0) - mat(2,2) ) * 2;

      quat.x() = (mat(1,0) + mat(0,1) ) / s;
      quat.y() = 0.5 / s;
      quat.z() = (mat(2,1) + mat(1,2) ) / s;
      quat.w() = (mat(2,0) + mat(0,2) ) / s;
    }
  else if (mat(2,2) > mat(1,1) && mat(2,2) > mat(3,3))
    {
      double s  = sqrt( 1.0 + mat(2,2) - mat(0,0) - mat(1,1) ) * 2;

      quat.x() = (mat(2,0) + mat(0,2) ) / s;
      quat.y() = (mat(2,1) + mat(1,2) ) / s;
      quat.z() = 0.5 / s;
      quat.w() = (mat(1,0) + mat(0,1) ) / s;
    }

  quat.getRotate(angle,axis);
}

//----------------------------------------------------------
bool OsgTransformObject::hasTopologyChanged() const
//----------------------------------------------------------
{
  return topo_state_.hasChanged(topo_state_orig);
}


//----------------------------------------------------------
void OsgTransformObject::clearTopologyState()
//----------------------------------------------------------
{
  topo_state_.clear(topo_state_orig);
}

 
//----------------------------------------------------------
bool OsgTransformObject::hasGeometryChanged() const
//----------------------------------------------------------
{
  return geom_state_.hasChanged(geom_state_orig);
}

//----------------------------------------------------------
void OsgTransformObject::clearGeometryState()
//----------------------------------------------------------
{
  geom_state_.clear(geom_state_orig);
}

//----------------------------------------------------------
ObjectState &OsgTransformObject::drawingState()
//----------------------------------------------------------
{
  return draw_state_;
}

//----------------------------------------------------------
bool OsgTransformObject::hasDrawingChanged() const
//----------------------------------------------------------
{
  return draw_state_.hasChanged(draw_state_orig);
}

//----------------------------------------------------------
void OsgTransformObject::clearDrawingState()
//----------------------------------------------------------
{
  draw_state_.clear(draw_state_orig);
}
/*
//----------------------------------------------------------
void OsgTransformObject::setState(SimResState *state)
//----------------------------------------------------------
{
  simres_state_ = state;
}
//----------------------------------------------------------
SimResState *OsgTransformObject::getState()
//----------------------------------------------------------
{
  return simres_state_.get();
}
*/
