#ifndef Util_SimResState_h_Included          // -*- C++ -*-
#define Util_SimResState_h_Included

#include <QObject>
#include <QColor>
#include <osg/Referenced>

/**
 * A SimResState stores the states of a GeometryNode that may vary 
 * during an animation.
 *
 *
 * \author Trond Vidar Stensby
 * \author Benjamin Kehlet
 * \date 2010-03-27
 */

class SimResState : public QObject, public osg::Referenced
{
  Q_OBJECT;
  
public:    
  enum StateId {
    RenderMode=1,
    SortMode,
    CullMode,
    Highlight,
    TwoSidedLighting,
    RelativeFieldTime,
    Transparency,
    DiffuseIntensity,
    AmbientIntensity,
    EmissionIntensity,
    SpecularIntensity,
    Shininess,
    BaseColor,
    BackColor,
    HighlightColor,
    WireframeColor,
    PointSize,
    LineWidth,
    Contour,
    ContourNb,
    BoundingSphere,
    TriangleNormals,
    Axes,
    Plane
  };

  enum RENDER_MODE
  {
    SMOOTH = 1,
    WIREFRAME,
    FLAT,
    WIRE_AND_FLAT,
    POINTS,
    WIRE_AND_SMOOTH
  };
    
  /** Uncommented */
  SimResState();
    
  /** Uncommented */
  SimResState(const SimResState& path);

  /** Uncommented */
  ~SimResState();

  // Get/set methods for properties

  /** Set render mode. */
  bool setRenderMode(int mode);

  /** Get render mode. */
  int getRenderMode() const;

  /** Set sort mode. */
  bool setSortMode(int mode);

  /** Get sort mode. */
  const int &getSortMode() const;

  /** Set cull mode. */
  int setCullMode(int mode);

  /** Get cull mode. */
  const int &getCullMode() const;

  /** Set highlight on/off. */
  bool setHighlight(bool mode);

  /** Get highlight on/off */
  const bool &getHighlight() const;

  /** Set axis on/off */
  bool setShowAxes(bool mode);

  /** Get show axes on/off **/
  bool getShowAxes() const;

  /** Set ground plane on/off */
  bool setShowPlane(bool mode);

  bool getShowPlane() const;

  /** Set two sided lighting. */
  bool setTwoSidedLighting(bool mode);

  /** Get two sided lighting. */
  const bool &getTwoSidedLighting() const;

  /** Set transparency. */
  bool setTransparency(double t);

  /** Get transparency. */
  const double &getTransparency() const;

  /** Set intensity of diffuse color. */
  bool setDiffuseIntensity(double value);

  /** Get intensity of diffuse color. */
  const double &getDiffuseIntensity() const;

  /** Set intensity of ambient color. */
  bool setAmbientIntensity(double value);

  /** Get intensity of ambient color. */
  const double &getAmbientIntensity() const;
    
  /** Set intensity of emission color. */
  bool setEmissionIntensity(double value);

  /** Get intensity of emission color. */
  const double &getEmissionIntensity() const;

  /** Set intensity of specular color. */
  bool setSpecularIntensity(double value);

  /** Get intensity of specular color. */
  const double &getSpecularIntensity() const;

  /** Set intensity of shininess color. */
  bool setShininess(double value);

  /** Get intensity of shininess color. */
  const double &getShininess() const;

  /** Set base (front) color of object. */
  bool setBaseColor(QColor color);

  /** Get base (front) color of object. */
  const QColor &getBaseColor() const;

  /** Set back color of object. */
  bool setBackColor(QColor color);

  /** Get back color of object. */
  const QColor &getBackColor() const;

  /** Set highlight color of object. */
  bool setHighlightColor(QColor color);

  /** Get highlight color of object. */
  const QColor &getHighlightColor() const;

  /** Set wireframe color of object. */
  bool setWireframeColor(QColor color);

  /** Get wireframe color of object. */
  const QColor &getWireframeColor() const;

  /** Set point width when drawing points */
  bool setPointSize(double line_width);
    
  /** Get point width when drawing points */
  const double &getPointSize() const;
    
  /** Set line width when drawing lines */
  bool setLineWidth(double line_width);

  /** Get line width when drawing lines */
  const double &getLineWidth() const;

  /** Show or hide contours */
  bool setContour(const bool onoff);

  /** True if show contours */
  const bool &getContour() const;

  /** Sets number of parallel contours to show */
  bool setContourNb(const int nb);

  /** Returns number of parallel contours to show */
  const int &getContourNb() const;

  /** Shows or hide the bounding sphere */
  bool showBoundingSphere(const bool onoff);

  /** True if the bounding sphere is shown */
  const bool &isBoundingSphereShown() const;

  /** Shows or hide the triangle normals */
  bool showTriangleNormals(const bool onoff);

  /** True if the triangle normals are shown */
  const bool &areTriangleNormalsShown() const;

signals:
  /** Emit signal when a state is changed. */
  void stateChanged(int id);

protected:

  
  int render_mode_;
  
  int sort_mode_;
  
  int culling_;
  
  bool highlight_;
  bool two_sided_;
  
  double transparency_;
  
  double diffuse_intensity_;
  double ambient_intensity_;
  double emission_intensity_;
  double specular_intensity_;
  double shininess_;
  
  QColor base_color_;
  QColor back_color_;
  QColor highlight_color_;
  QColor wireframe_color_;
  
  double point_size_;
  double line_width_;
  
  bool contour_;
  int  contour_nb_;
  
  bool show_bound_;
  bool show_triangle_normal_;
  bool show_axes_;
  bool show_plane_;
};

#endif // Util_SimResState_h_Included
